import numpy as np
import pandas as pd
cimport numpy as np
from scipy.special.cython_special import ndtri, stdtrit, gamma
from collections import deque

from qubx.core.series cimport TimeSeries, Indicator, IndicatorOHLC, RollingSum, nans, OHLCV, Bar
from qubx.core.utils import time_to_str
from qubx.pandaz.utils import scols, srows


cdef extern from "math.h":
    float INFINITY


cdef class Sma(Indicator):
    """
    Simple Moving Average indicator
    """

    def __init__(self, str name, TimeSeries series, int period):
        self.period = period
        self.summator = RollingSum(period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        cdef double r = self.summator.update(value, new_item_started)
        return np.nan if self.summator.is_init_stage else r / self.period


def sma(series:TimeSeries, period: int): 
    return Sma.wrap(series, period)


cdef class Ema(Indicator):
    """
    Exponential moving average
    """

    def __init__(self, str name, TimeSeries series, int period, init_mean=True):
        self.period = period

        # when it's required to initialize this ema by mean on first period
        self.init_mean = init_mean
        if init_mean:
            self.__s = nans(period)
            self.__i = 0

        self._init_stage = 1
        self.alpha = 2.0 / (1.0 + period)
        self.alpha_1 = (1 - self.alpha)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        cdef int prev_bar_idx = 0 if new_item_started else 1 

        if self._init_stage:
            if np.isnan(value): return np.nan

            if new_item_started:
                self.__i += 1
                if self.__i > self.period - 1:
                    self._init_stage = False
                    return self.alpha * value + self.alpha_1 * self[prev_bar_idx]

            if self.__i == self.period - 1:
                self.__s[self.__i] = value 
                return np.nansum(self.__s) / self.period

            self.__s[self.__i] = value 
            return np.nan

        if len(self) == 0:
            return value

        return self.alpha * value + self.alpha_1 * self[prev_bar_idx]


def ema(series:TimeSeries, period: int, init_mean: bool = True):
    return Ema.wrap(series, period, init_mean=init_mean)


cdef class Tema(Indicator):

    def __init__(self, str name, TimeSeries series, int period, init_mean=True):
        self.period = period
        self.init_mean = init_mean
        self.ser0 = TimeSeries('ser0', series.timeframe, series.max_series_length)
        self.ema1 = ema(self.ser0, period, init_mean)
        self.ema2 = ema(self.ema1, period, init_mean)
        self.ema3 = ema(self.ema2, period, init_mean)
        super().__init__(name, series)
        
    cpdef double calculate(self, long long time, double value, short new_item_started):
        self.ser0.update(time, value)
        return 3 * self.ema1[0] - 3 * self.ema2[0] + self.ema3[0]


def tema(series:TimeSeries, period: int, init_mean: bool = True):
    return Tema.wrap(series, period, init_mean=init_mean)


cdef class Dema(Indicator):

    def __init__(self, str name, TimeSeries series, int period, init_mean=True):
        self.period = period
        self.init_mean = init_mean
        self.ser0 = TimeSeries('ser0', series.timeframe, series.max_series_length)
        self.ema1 = ema(self.ser0, period, init_mean)
        self.ema2 = ema(self.ema1, period, init_mean)
        super().__init__(name, series)
        
    cpdef double calculate(self, long long time, double value, short new_item_started):
        self.ser0.update(time, value)
        return 2 * self.ema1[0] - self.ema2[0]


def dema(series:TimeSeries, period: int, init_mean: bool = True):
    return Dema.wrap(series, period, init_mean=init_mean)


cdef class Kama(Indicator):
    # cdef int period
    # cdef int fast_span
    # cdef int slow_span
    # cdef double _S1 
    # cdef double _K1 
    # cdef _x_past
    # cdef RollingSum summator

    def __init__(self, str name, TimeSeries series, int period, int fast_span=2, int slow_span=30):
        self.period = period
        self.fast_span = fast_span
        self.slow_span = slow_span
        self._S1 = 2.0 / (slow_span + 1)
        self._K1 = 2.0 / (fast_span + 1) - self._S1
        self._x_past = deque(nans(period+1), period+1)
        self.summator = RollingSum(period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        if new_item_started:
            self._x_past.append(value)
        else:
            self._x_past[-1] = value

        cdef double rs = self.summator.update(abs(value - self._x_past[-2]), new_item_started)
        cdef double er = (abs(value - self._x_past[0]) / rs) if rs != 0.0 else 1.0
        cdef double sc = (er * self._K1 + self._S1) ** 2

        if self.summator.is_init_stage:
            if not np.isnan(self._x_past[1]):
                return value
            return np.nan

        return sc * value + (1 - sc) * self[0 if new_item_started else 1]


def kama(series:TimeSeries, period: int, fast_span:int=2, slow_span:int=30):
    return Kama.wrap(series, period, fast_span, slow_span)


cdef class Highest(Indicator):

    def __init__(self, str name, TimeSeries series, int period):
        self.period = period
        self.queue = deque([np.nan] * period, maxlen=period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        """
        Not a most effictive algo but simplest and can handle updated last value
        """
        cdef float r = np.nan

        if not np.isnan(value):
            if new_item_started:
                self.queue.append(value)
            else:
                self.queue[-1] = value

        if not np.isnan(self.queue[0]):
            r = max(self.queue) 

        return r


def highest(series:TimeSeries, period:int):
    return Highest.wrap(series, period)


cdef class Lowest(Indicator):

    def __init__(self, str name, TimeSeries series, int period):
        self.period = period
        self.queue = deque([np.nan] * period, maxlen=period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        """
        Not a most effictive algo but simplest and can handle updated last value
        """
        cdef float r = np.nan

        if not np.isnan(value):
            if new_item_started:
                self.queue.append(value)
            else:
                self.queue[-1] = value

        if not np.isnan(self.queue[0]):
            r = min(self.queue) 

        return r


def lowest(series:TimeSeries, period:int):
    return Lowest.wrap(series, period)


cdef class Std(Indicator):
    """
    Streaming Standard Deviation indicator
    """

    def __init__(self, str name, TimeSeries series, int period):
        self.period = period
        self.rolling_sum = RollingSum(period)
        self.variance_sum = RollingSum(period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        # Update the rolling sum with the new value
        cdef double _sum = self.rolling_sum.update(value, new_item_started)

        # If we're still in the initialization stage, return NaN
        if self.rolling_sum.is_init_stage:
            return np.nan

        # Calculate the mean from the rolling sum
        cdef double _mean = _sum / self.period

        # Update the variance sum with the squared deviation from the mean
        cdef double _var_sum = self.variance_sum.update((value - _mean) ** 2, new_item_started)

        # If the variance sum is still in the initialization stage, return NaN
        if self.variance_sum.is_init_stage:
            return np.nan

        # Return the square root of the variance (standard deviation)
        return np.sqrt(_var_sum / self.period)


def std(series: TimeSeries, period: int, mean=0):
    return Std.wrap(series, period)


cdef double norm_pdf(double x):
    return np.exp(-x ** 2 / 2) / np.sqrt(2 * np.pi)


cdef double lognorm_pdf(double x, double s):
    return np.exp(-np.log(x) ** 2 / (2 * s ** 2)) / (x * s * np.sqrt(2 * np.pi))


cdef double student_t_pdf(double x, double df):
    """Compute the PDF of the Student's t-distribution."""
    gamma_df = gamma(df / 2.0)
    gamma_df_plus_1 = gamma((df + 1) / 2.0)
    
    # Normalization constant
    normalization = gamma_df_plus_1 / (np.sqrt(df * np.pi) * gamma_df)
    
    # PDF calculation
    term = (1 + (x ** 2) / df) ** (-(df + 1) / 2.0)
    pdf_value = normalization * term
    
    return pdf_value


cdef class Pewma(Indicator):

    def __init__(self, str name, TimeSeries series, double alpha, double beta, int T):
        self.alpha = alpha 
        self.beta = beta
        self.T = T

        # - local variables
        self._i = 0
        self.std = TimeSeries('std', series.timeframe, series.max_series_length)
        super().__init__(name, series)

    def _store(self):
        self.mean = self._mean
        self.vstd = self._vstd
        self.var = self._var

    def _restore(self):
        self._mean = self.mean
        self._vstd = self.vstd
        self._var = self.var

    def _get_alpha(self, p_t):
        if self._i - 1 > self.T:
            return self.alpha * (1.0 - self.beta * p_t)
        return 1.0 - 1.0 / self._i

    cpdef double calculate(self, long long time, double x, short new_item_started):
        cdef double diff, p_t, a_t, incr

        if len(self.series) <= 1:
            self._mean = x
            self._vstd = 0.0
            self._var = 0.0
            self._store()
            self.std.update(time, self.vstd)
            return self.mean

        if new_item_started:
            self._i += 1
            self._restore()
        else:
            self._store()

        diff = x - self.mean
        # prob of observing diff
        p_t = norm_pdf(diff / self.vstd) if self.vstd != 0.0 else 0.0  

        # weight to give to this point
        a_t = self._get_alpha(p_t)  
        incr = (1.0 - a_t) * diff
        self.mean += incr
        self.var = a_t * (self.var + diff * incr)
        self.vstd = np.sqrt(self.var)
        self.std.update(time, self.vstd)

        return self.mean


def pewma(series:TimeSeries, alpha: float, beta: float, T:int=30):
    """
    Implementation of probabilistic exponential weighted ma (https://sci-hub.shop/10.1109/SSP.2012.6319708)
    See pandas version here: qubx.pandaz.ta::pwma 
    """
    return Pewma.wrap(series, alpha, beta, T)


cdef class PewmaOutliersDetector(Indicator):

    def __init__(
        self,
        str name,
        TimeSeries series,
        double alpha,
        double beta,
        int T,
        double threshold,
        str dist = "normal",
        double student_t_df = 3.0
    ):
        self.alpha = alpha 
        self.beta = beta
        self.T = T
        self.threshold = threshold
        self.dist = dist
        self.student_t_df = student_t_df

        # - series
        self.upper = TimeSeries('uba', series.timeframe, series.max_series_length)
        self.lower = TimeSeries('lba', series.timeframe, series.max_series_length)
        self.std = TimeSeries('std', series.timeframe, series.max_series_length)
        self.outliers = TimeSeries('outliers', series.timeframe, series.max_series_length)

        # - local variables
        self._i = 0
        self._z_thr = self._get_z_thr()

        super().__init__(name, series)

    def _store(self):
        self.mean = self._mean
        self.vstd = self._vstd
        self.variance = self._variance

    def _restore(self):
        self._mean = self.mean
        self._vstd = self.vstd
        self._variance = self.variance
    
    cdef double _get_z_thr(self):
        if self.dist == 'normal':
            return ndtri(1 - self.threshold / 2)
        elif self.dist == 'student_t':
            return stdtrit(self.student_t_df, 1 - self.threshold / 2)
        else:
            raise ValueError('Invalid distribution type')

    cdef double _get_alpha(self, double p_t):
        if self._i + 1 >= self.T:
            return self.alpha * (1.0 - self.beta * p_t)
        return 1.0 - 1.0 / (self._i + 1.0)

    cdef double _get_mean(self, double x, double alpha_t):
        return alpha_t * self.mean + (1.0 - alpha_t) * x

    cdef double _get_variance(self, double x, double alpha_t):
        return alpha_t * self.variance + (1.0 - alpha_t) * np.square(x)

    cdef double _get_std(self, double variance, double mean):
        return np.sqrt(max(variance - np.square(mean), 0.0))

    cdef double _get_p(self, double x):
        cdef double z_t = ((x - self.mean) / self.vstd) if (self.vstd != 0 and not np.isnan(x)) else 0.0
        if self.dist == 'normal':
            p_t = norm_pdf(z_t)
        elif self.dist == 'student_t':
            p_t = student_t_pdf(z_t, self.student_t_df)
        # elif self.dist == 'cauchy':
        #     p_t = (1 / (np.pi * (1 + np.square(z_t))))
        else:
            raise ValueError('Invalid distribution type')
        return p_t

    cpdef double calculate(self, long long time, double x, short new_item_started):
        # - first bar - just use it as initial value
        if len(self.series) <= 1:
            self._mean = x
            self._variance = x ** 2
            self._vstd = 0.0
            self._store()
            self.std.update(time, self.vstd)
            self.upper.update(time, x)
            self.lower.update(time, x)
            return self._mean

        # - new bar is started use n-1 values for calculate innovations
        if new_item_started:
            self._i += 1
            self._restore()
        else:
            self._store()

        cdef double p_t = self._get_p(x)
        cdef double alpha_t = self._get_alpha(p_t)
        self.mean = self._get_mean(x, alpha_t)
        self.variance = self._get_variance(x, alpha_t)
        self.vstd = self._get_std(self.variance, self.mean)
        cdef double ub = self.mean + self._z_thr * self.vstd
        cdef double lb = self.mean - self._z_thr * self.vstd

        self.upper.update(time, ub)
        self.lower.update(time, lb)
        self.std.update(time, self.vstd)

        # - check if it's outlier
        if p_t < self.threshold:
            self.outliers.update(time, x)
        else:
            self.outliers.update(time, np.nan)
        return self.mean


def pewma_outliers_detector(
    series: TimeSeries,
    alpha: float,
    beta: float,
    T:int=30,
    threshold=0.05,
    dist: str = "normal",
    **kwargs
):
    """
    Outliers detector based on pwma
    """
    return PewmaOutliersDetector.wrap(series, alpha, beta, T, threshold, dist=dist, **kwargs)


cdef class Psar(IndicatorOHLC):

    def __init__(self, name, series, iaf, maxaf):
        self.iaf = iaf
        self.maxaf = maxaf
        self.upper = TimeSeries('upper', series.timeframe, series.max_series_length)
        self.lower = TimeSeries('lower', series.timeframe, series.max_series_length)
        super().__init__(name, series)

    cdef _store(self):
        self.bull = self._bull
        self.af = self._af
        self.psar = self._psar
        self.lp = self._lp
        self.hp = self._hp

    cdef _restore(self):
        self._bull = self.bull
        self._af = self.af
        self._psar = self.psar
        self._lp = self.lp
        self._hp = self.hp

    cpdef double calculate(self, long long time, Bar bar, short new_item_started):
        cdef short reverse = 1

        if len(self.series) <= 2:
            self._bull = 1
            self._af = self.iaf
            self._psar = bar.close

            if len(self.series) == 1:
                self._lp = bar.low
                self._hp = bar.high
            self._store()
            return self._psar

        if not new_item_started:
            self._store()
        else:
            self._restore()

        bar1 = self.series[1]
        bar2 = self.series[2]
        cdef double h0 = bar.high
        cdef double l0 = bar.low
        cdef double h1 = bar1.high
        cdef double l1 = bar1.low
        cdef double h2 = bar2.high
        cdef double l2 = bar2.low

        if self.bull:
            self.psar += self.af * (self.hp - self.psar)
        else:
            self.psar += self.af * (self.lp - self.psar)

        reverse = 0
        if self.bull:
            if l0 < self.psar:
                self.bull = 0
                reverse = 1
                self.psar = self.hp
                self.lp = l0
                self.af = self.iaf
        else:
            if h0 > self.psar:
                self.bull = 1
                reverse = 1
                self.psar = self.lp
                self.hp = h0
                self.af = self.iaf

        if not reverse:
            if self.bull:
                if h0 > self.hp:
                    self.hp = h0
                    self.af = min(self.af + self.iaf, self.maxaf)
                if l1 < self.psar:
                    self.psar = l1
                if l2 < self.psar:
                    self.psar = l2
            else:
                if l0 < self.lp:
                    self.lp = l0
                    self.af = min(self.af + self.iaf, self.maxaf)
                if h1 > self.psar:
                    self.psar = h1
                if h2 > self.psar:
                    self.psar = h2

        if self.bull:
            self.lower.update(time, self.psar)
            self.upper.update(time, np.nan)
        else:
            self.upper.update(time, self.psar)
            self.lower.update(time, np.nan)

        return self.psar


def psar(series: OHLCV, iaf: float=0.02, maxaf: float=0.2):
    if not isinstance(series, OHLCV):
        raise ValueError('Series must be OHLCV !')

    return Psar.wrap(series, iaf, maxaf)


# List of smoothing functions
_smoothers = {f.__name__: f for f in [pewma, ema, sma, kama, tema, dema]}


def smooth(TimeSeries series, str smoother, *args, **kwargs) -> Indicator:
    """
    Handy utility function to smooth series
    """
    _sfn = _smoothers.get(smoother)
    if _sfn is None:
        raise ValueError(f"Smoother {smoother} not found!")
    return _sfn(series, *args, **kwargs)


cdef class Atr(IndicatorOHLC):

    def __init__(self, str name, OHLCV series, int period, str smoother, short percentage):
        self.percentage = percentage
        self.tr = TimeSeries("tr", series.timeframe, series.max_series_length)
        self.ma = smooth(self.tr, smoother, period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, Bar bar, short new_item_started):
        if len(self.series) <= 1:
            return np.nan

        cdef double c1 = self.series[1].close
        cdef double h_l = abs(bar.high - bar.low)
        cdef double h_pc = abs(bar.high - c1)
        cdef double l_pc = abs(bar.low - c1)
        self.tr.update(time, max(h_l, h_pc, l_pc))
        return (100 * self.ma[0] / c1) if self.percentage else self.ma[0]


def atr(series: OHLCV, period: int = 14, smoother="sma", percentage: bool = False):
    if not isinstance(series, OHLCV):
        raise ValueError("Series must be OHLCV !")
    return Atr.wrap(series, period, smoother, percentage)


cdef class Zscore(Indicator):
    """
    Z-score normalization using rolling SMA and Std
    """

    def __init__(self, str name, TimeSeries series, int period, str smoother):
        self.tr = TimeSeries("tr", series.timeframe, series.max_series_length)
        self.ma = smooth(self.tr, smoother, period)
        self.std = std(self.tr, period)
        super().__init__(name, series)

    cpdef double calculate(self, long long time, double value, short new_item_started):
        self.tr.update(time, value)
        if len(self.ma) < 1 or len(self.std) < 1 or np.isnan(self.ma[0]) or np.isnan(self.std[0]) or self.std[0] == 0:
            return np.nan
        return (value - self.ma[0]) / self.std[0]


def zscore(series: TimeSeries, period: int = 20, smoother="sma"):
    return Zscore.wrap(series, period, smoother)


cdef class Swings(IndicatorOHLC):

    def __init__(self, str name, OHLCV series, trend_indicator, **indicator_args):
        self.base = OHLCV("base", series.timeframe, series.max_series_length)
        self.trend = trend_indicator(self.base, **indicator_args)

        self.tops = TimeSeries("tops", series.timeframe, series.max_series_length)
        self.tops_detection_lag = TimeSeries("tops_lag", series.timeframe, series.max_series_length)

        self.bottoms = TimeSeries("bottoms", series.timeframe, series.max_series_length)
        self.bottoms_detection_lag = TimeSeries("bottoms_lag", series.timeframe, series.max_series_length)

        self.middles = TimeSeries("middles", series.timeframe, series.max_series_length)
        self.deltas = TimeSeries("deltas", series.timeframe, series.max_series_length)

        # - store parameters for copying
        self._trend_indicator = trend_indicator
        self._indicator_args = indicator_args

        self._min_l = +np.inf
        self._max_h = -np.inf
        self._max_t = 0
        self._min_t = 0
        super().__init__(name, series)

    cpdef double calculate(self, long long time, Bar bar, short new_item_started):
        self.base.update_by_bar(time, bar.open, bar.high, bar.low, bar.close, bar.volume)
        cdef int _t = 0

        if len(self.trend.upper) > 0:
            _u = self.trend.upper[0]
            _d = self.trend.lower[0]

            if not np.isnan(_u):
                if self._max_t > 0:
                    self.tops.update(self._max_t, self._max_h)
                    self.tops_detection_lag.update(self._max_t, time - self._max_t)
                    if len(self.bottoms) > 0:
                        self.middles.update(time, (self.tops[0] + self.bottoms[0]) / 2)
                        self.deltas.update(time, self.tops[0] - self.bottoms[0])

                if bar.low <= self._min_l:
                    self._min_l = bar.low
                    self._min_t = time

                self._max_h = -np.inf
                self._max_t = 0
                _t = -1
            elif not np.isnan(_d):
                if self._min_t > 0:
                    self.bottoms.update(self._min_t, self._min_l)
                    self.bottoms_detection_lag.update(self._min_t, time - self._min_t)
                    if len(self.tops) > 0:
                        self.middles.update(time, (self.tops[0] + self.bottoms[0]) / 2)
                        self.deltas.update(time, self.tops[0] - self.bottoms[0])

                if bar.high >= self._max_h:
                    self._max_h = bar.high
                    self._max_t = time

                self._min_l = +np.inf
                self._min_t = 0
                _t = +1

        return _t

    def get_current_trend_end(self):
        if np.isfinite(self._min_l):
            return pd.Timestamp(self._min_t, 'ns'), self._min_l
        elif np.isfinite(self._max_h):
            return pd.Timestamp(self._max_t, 'ns'), self._max_h
        return (None, None)

    def copy(self, int start, int stop):
        n_ts = Swings(self.name, OHLCV("base", self.series.timeframe), self._trend_indicator, **self._indicator_args)

        # - copy main series
        for i in range(start, stop):
            n_ts._add_new_item(self.times.values[i], self.values.values[i])
            n_ts.trend._add_new_item(self.trend.times.values[i], self.trend.values.values[i])

        # - copy internal series
        (
            n_ts.tops, 
            n_ts.tops_detection_lag,
            n_ts.bottoms,
            n_ts.bottoms_detection_lag,
            n_ts.middles,
            n_ts.deltas
        ) = self._copy_internal_series(start, stop, 
            self.tops, 
            self.tops_detection_lag,
            self.bottoms,
            self.bottoms_detection_lag,
            self.middles,
            self.deltas
        )

        return n_ts

    def pd(self) -> pd.DataFrame:
        _t, _d = self.get_current_trend_end()
        tps, bts = self.tops.pd(), self.bottoms.pd()
        tpl, btl = self.tops_detection_lag.pd(), self.bottoms_detection_lag.pd()
        if _t is not None:
            if bts.index[-1] < tps.index[-1]:
                bts = srows(bts, pd.Series({_t: _d}))
                btl = srows(btl, pd.Series({_t: 0}))  # last lag is 0
            else:
                tps = srows(tps, pd.Series({_t: _d}))
                tpl = srows(tpl, pd.Series({_t: 0})) # last lag is 0

        # - convert tpl / btl to timedeltas
        tpl, btl = tpl.apply(lambda x: pd.Timedelta(x, unit='ns')), btl.apply(lambda x: pd.Timedelta(x, unit='ns'))

        eid = pd.Series(tps.index, tps.index)
        mx = scols(bts, tps, eid, names=["start_price", "end_price", "end"])
        dt = scols(mx["start_price"], mx["end_price"].shift(-1), mx["end"].shift(-1))  # .dropna()
        dt = dt.assign(
            delta = dt["end_price"] - dt["start_price"], 
            spotted = pd.Series(bts.index + btl, bts.index)
        )

        eid = pd.Series(bts.index, bts.index)
        mx = scols(tps, bts, eid, names=["start_price", "end_price", "end"])
        ut = scols(mx["start_price"], mx["end_price"].shift(-1), mx["end"].shift(-1))  # .dropna()
        ut = ut.assign(
            delta = ut["end_price"] - ut["start_price"], 
            spotted = pd.Series(tps.index + tpl, tps.index)
        )

        return scols(ut, dt, keys=["DownTrends", "UpTrends"])


def swings(series: OHLCV, trend_indicator, **indicator_args):
    """
    Swing detector based on provided trend indicator.
    """
    if not isinstance(series, OHLCV):
        raise ValueError("Series must be OHLCV !")
    return Swings.wrap(series, trend_indicator, **indicator_args)


from collections import defaultdict
from typing import Any, Dict, Optional

import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from qubx import logger
from qubx.backtester.simulated_data import IterableSimulationData
from qubx.core.basics import (
    CtrlChannel,
    DataType,
    Instrument,
    TimestampedDict,
)
from qubx.core.exceptions import SimulationError
from qubx.core.helpers import BasicScheduler
from qubx.core.interfaces import IDataProvider
from qubx.core.series import Bar, Quote, time_as_nsec
from qubx.data.readers import AsDict, DataReader
from qubx.utils.time import infer_series_frequency

from .account import SimulatedAccountProcessor
from .utils import SimulatedTimeProvider


class SimulatedDataProvider(IDataProvider):
    time_provider: SimulatedTimeProvider
    channel: CtrlChannel

    _scheduler: BasicScheduler
    _account: SimulatedAccountProcessor
    _last_quotes: Dict[Instrument, Optional[Quote]]
    _readers: dict[str, DataReader]
    _scheduler: BasicScheduler
    _pregenerated_signals: dict[Instrument, pd.Series | pd.DataFrame]
    _to_process: dict[Instrument, list]
    _data_source: IterableSimulationData
    _open_close_time_indent_ns: int

    def __init__(
        self,
        exchange_id: str,
        channel: CtrlChannel,
        scheduler: BasicScheduler,
        time_provider: SimulatedTimeProvider,
        account: SimulatedAccountProcessor,
        readers: dict[str, DataReader],
        open_close_time_indent_secs=1,
    ):
        self.channel = channel
        self.time_provider = time_provider
        self._exchange_id = exchange_id
        self._scheduler = scheduler
        self._account = account
        self._readers = readers

        # - create exchange's instance
        self._last_quotes = defaultdict(lambda: None)

        # - pregenerated signals storage
        self._pregenerated_signals = dict()
        self._to_process = {}

        # - simulation data source
        self._data_source = IterableSimulationData(
            self._readers, open_close_time_indent_secs=open_close_time_indent_secs
        )
        self._open_close_time_indent_ns = open_close_time_indent_secs * 1_000_000_000  # convert seconds to nanoseconds

        logger.info(f"{self.__class__.__name__}.{exchange_id} is initialized")

    def run(
        self,
        start: str | pd.Timestamp,
        end: str | pd.Timestamp,
        silent: bool = False,
    ) -> None:
        logger.info(f"{self.__class__.__name__} ::: Simulation started at {start} :::")

        if self._pregenerated_signals:
            self._prepare_generated_signals(start, end)
            _run = self._process_generated_signals
        else:
            _run = self._process_strategy

        start, end = pd.Timestamp(start), pd.Timestamp(end)
        total_duration = end - start
        update_delta = total_duration / 100
        prev_dt = pd.Timestamp(start)

        # - date iteration
        qiter = self._data_source.create_iterable(start, end)
        if silent:
            for instrument, data_type, event, is_hist in qiter:
                if not _run(instrument, data_type, event, is_hist):
                    break
        else:
            _p = 0
            with tqdm(total=100, desc="Simulating", unit="%", leave=False) as pbar:
                for instrument, data_type, event, is_hist in qiter:
                    if not _run(instrument, data_type, event, is_hist):
                        break
                    dt = pd.Timestamp(event.time)
                    # update only if date has changed
                    if dt - prev_dt > update_delta:
                        _p += 1
                        pbar.n = _p
                        pbar.refresh()
                        prev_dt = dt
                pbar.n = 100
                pbar.refresh()

        logger.info(f"{self.__class__.__name__} ::: Simulation finished at {end} :::")

    def set_generated_signals(self, signals: pd.Series | pd.DataFrame):
        logger.debug(
            f"[<y>{self.__class__.__name__}</y>] :: Using pre-generated signals:\n {str(signals.count()).strip('ndtype: int64')}"
        )
        # - sanity check
        signals.index = pd.DatetimeIndex(signals.index)

        if isinstance(signals, pd.Series):
            self._pregenerated_signals[str(signals.name)] = signals  # type: ignore

        elif isinstance(signals, pd.DataFrame):
            for col in signals.columns:
                self._pregenerated_signals[col] = signals[col]  # type: ignore
        else:
            raise ValueError("Invalid signals or strategy configuration")

    @property
    def is_simulation(self) -> bool:
        return True

    def subscribe(self, subscription_type: str, instruments: set[Instrument], reset: bool) -> None:
        _new_instr = [i for i in instruments if not self.has_subscription(i, subscription_type)]
        self._data_source.add_instruments_for_subscription(subscription_type, list(instruments))

        # - provide historical data and last quote for subscribed instruments
        for i in _new_instr:
            h_data = self._data_source.peek_historical_data(i, subscription_type)
            if h_data:
                # _s_type = DataType.from_str(subscription_type)[0]
                last_update = h_data[-1]
                if last_quote := self._account.emulate_quote_from_data(i, last_update.time, last_update):  # type: ignore
                    # - send historical data to the channel
                    self.channel.send((i, subscription_type, h_data, True))

                    # - set last quote
                    self._last_quotes[i] = last_quote

                    # - also need to pass this quote to OME !
                    self._account._process_new_quote(i, last_quote)

                    logger.debug(f" | subscribed {subscription_type} {i} -> {last_quote}")

    def unsubscribe(self, subscription_type: str, instruments: set[Instrument] | Instrument | None = None) -> None:
        # logger.debug(f" | unsubscribe: {subscription_type} -> {instruments}")
        if instruments is not None:
            self._data_source.remove_instruments_from_subscription(
                subscription_type, [instruments] if isinstance(instruments, Instrument) else list(instruments)
            )

    def has_subscription(self, instrument: Instrument, subscription_type: str) -> bool:
        return self._data_source.has_subscription(instrument, subscription_type)

    def get_subscriptions(self, instrument: Instrument) -> list[str]:
        _s_lst = self._data_source.get_subscriptions_for_instrument(instrument)
        # logger.debug(f" | get_subscriptions {instrument} -> {_s_lst}")
        return _s_lst

    def get_subscribed_instruments(self, subscription_type: str | None = None) -> list[Instrument]:
        _in_lst = self._data_source.get_instruments_for_subscription(subscription_type or DataType.ALL)
        # logger.debug(f" | get_subscribed_instruments {subscription_type} -> {_in_lst}")
        return _in_lst

    def warmup(self, configs: dict[tuple[str, Instrument], str]) -> None:
        for si, warm_period in configs.items():
            logger.debug(f" | Warming up {si} -> {warm_period}")
            self._data_source.set_warmup_period(si[0], warm_period)

    def get_ohlc(self, instrument: Instrument, timeframe: str, nbarsback: int) -> list[Bar]:
        _reader = self._readers.get(DataType.OHLC)
        if _reader is None:
            logger.error(f"Reader for {DataType.OHLC} data not configured")
            return []

        start = pd.Timestamp(self.time_provider.time())
        end = start - nbarsback * (_timeframe := pd.Timedelta(timeframe))
        _spec = f"{instrument.exchange}:{instrument.symbol}"
        return self._convert_records_to_bars(
            _reader.read(data_id=_spec, start=start, stop=end, transform=AsDict()),  # type: ignore
            time_as_nsec(self.time_provider.time()),
            _timeframe.asm8.item(),
        )

    def get_quote(self, instrument: Instrument) -> Quote | None:
        return self._last_quotes[instrument]

    def close(self):
        pass

    def _prepare_generated_signals(self, start: str | pd.Timestamp, end: str | pd.Timestamp):
        for s, v in self._pregenerated_signals.items():
            _s_inst = None

            for i in self.get_subscribed_instruments():
                # - we can process series with variable id's if we can find some similar instrument
                if s == i.symbol or s == str(i) or s == f"{i.exchange}:{i.symbol}" or str(s) == str(i):
                    _start, _end = pd.Timestamp(start), pd.Timestamp(end)
                    _start_idx, _end_idx = v.index.get_indexer([_start, _end], method="ffill")
                    sel = v.iloc[max(_start_idx, 0) : _end_idx + 1]

                    # TODO: check if data has exec_price - it means we have deals
                    self._to_process[i] = list(zip(sel.index, sel.values))
                    _s_inst = i
                    break

            if _s_inst is None:
                logger.error(f"Can't find instrument for pregenerated signals with id '{s}'")
                raise SimulationError(f"Can't find instrument for pregenerated signals with id '{s}'")

    def _convert_records_to_bars(
        self, records: list[TimestampedDict], cut_time_ns: int, timeframe_ns: int
    ) -> list[Bar]:
        """
        Convert records to bars and we need to cut last bar up to the cut_time_ns
        """
        bars = []

        # - if no records, return empty list to avoid exception from infer_series_frequency
        if not records:
            return bars

        _data_tf = infer_series_frequency([r.time for r in records[:50]])
        timeframe_ns = _data_tf.item()

        if records is not None:
            for r in records:
                # _b_ts_0 = np.datetime64(r.time, "ns").item()
                _b_ts_0 = r.time
                _b_ts_1 = _b_ts_0 + timeframe_ns - self._open_close_time_indent_ns

                if _b_ts_0 <= cut_time_ns and cut_time_ns < _b_ts_1:
                    break

                bars.append(
                    Bar(
                        _b_ts_0, r.data["open"], r.data["high"], r.data["low"], r.data["close"], r.data.get("volume", 0)
                    )
                )

        return bars

    def _process_generated_signals(self, instrument: Instrument, data_type: str, data: Any, is_hist: bool) -> bool:
        cc = self.channel
        t = np.datetime64(data.time, "ns")

        if not is_hist:
            # - signals for this instrument
            sigs = self._to_process[instrument]

            while sigs and t >= (_signal_time := sigs[0][0].as_unit("ns").asm8):
                self.time_provider.set_time(_signal_time)
                cc.send((instrument, "event", {"order": sigs[0][1]}, False))
                sigs.pop(0)

            if q := self._account.emulate_quote_from_data(instrument, t, data):
                self._last_quotes[instrument] = q

        self.time_provider.set_time(t)
        cc.send((instrument, data_type, data, is_hist))

        return cc.control.is_set()

    def _process_strategy(self, instrument: Instrument, data_type: str, data: Any, is_hist: bool) -> bool:
        cc = self.channel
        t = np.datetime64(data.time, "ns")

        if not is_hist:
            if t >= (_next_exp_time := self._scheduler.next_expected_event_time()):
                # - we use exact event's time
                self.time_provider.set_time(_next_exp_time)
                self._scheduler.check_and_run_tasks()

            if q := self._account.emulate_quote_from_data(instrument, t, data):
                self._last_quotes[instrument] = q

        self.time_provider.set_time(t)
        cc.send((instrument, data_type, data, is_hist))

        return cc.control.is_set()

    def exchange(self) -> str:
        return self._exchange_id.upper()



import configparser
import dataclasses
import glob
import json
import os
import re
from datetime import datetime
from pathlib import Path

import stackprinter

from qubx import logger
from qubx.core.basics import ZERO_COSTS, AssetType, Instrument, MarketType, TransactionCostsCalculator
from qubx.utils.marketdata.dukas import SAMPLE_INSTRUMENTS
from qubx.utils.misc import get_local_qubx_folder, makedirs

_DEF_INSTRUMENTS_FOLDER = "instruments"
_DEF_FEES_FOLDER = "fees"


class _InstrumentEncoder(json.JSONEncoder):
    def default(self, obj):
        if dataclasses.is_dataclass(obj):
            return {k: v for k, v in dataclasses.asdict(obj).items() if not k.startswith("_")}
        if isinstance(obj, (datetime)):
            return obj.isoformat()
        return super().default(obj)


class _InstrumentDecoder(json.JSONDecoder):
    def decode(self, json_string):
        obj = super(_InstrumentDecoder, self).decode(json_string)
        if isinstance(obj, dict):
            # Convert delivery_date and onboard_date strings to datetime
            delivery_date = obj.get("delivery_date")
            onboard_date = obj.get("onboard_date")
            if delivery_date:
                obj["delivery_date"] = datetime.strptime(delivery_date.split(".")[0], "%Y-%m-%dT%H:%M:%S")
            if onboard_date:
                obj["onboard_date"] = datetime.strptime(onboard_date.split(".")[0], "%Y-%m-%dT%H:%M:%S")
            return Instrument(
                symbol=obj["symbol"],
                asset_type=AssetType[obj["asset_type"]],
                market_type=MarketType[obj["market_type"]],
                exchange=obj["exchange"],
                base=obj["base"],
                quote=obj["quote"],
                settle=obj["settle"],
                exchange_symbol=obj.get("exchange_symbol", obj["symbol"]),
                tick_size=float(obj["tick_size"]),
                lot_size=float(obj["lot_size"]),
                min_size=float(obj["min_size"]),
                min_notional=float(obj.get("min_notional", 0.0)),
                initial_margin=float(obj.get("initial_margin", 0.0)),
                maint_margin=float(obj.get("maint_margin", 0.0)),
                liquidation_fee=float(obj.get("liquidation_fee", 0.0)),
                contract_size=float(obj.get("contract_size", 1.0)),
                onboard_date=obj.get("onboard_date"),
                delivery_date=obj.get("delivery_date"),
            )
        elif isinstance(obj, list):
            return [self.decode(json.dumps(item)) for item in obj]
        return obj


class InstrumentsLookup:
    _lookup: dict[str, Instrument]
    _path: str

    def __init__(self, path: str = makedirs(get_local_qubx_folder(), _DEF_INSTRUMENTS_FOLDER)) -> None:
        self._path = path
        if not self.load():
            self.refresh()
        self.load()

    def load(self) -> bool:
        self._lookup = {}
        data_exists = False
        for fs in glob.glob(self._path + "/*.json"):
            try:
                with open(fs, "r") as f:
                    instrs: list[Instrument] = json.load(f, cls=_InstrumentDecoder)
                    for i in instrs:
                        self._lookup[f"{i.exchange}:{i.symbol}"] = i
                    data_exists = True
            except Exception as ex:
                stackprinter.show_current_exception()
                logger.warning(ex)

        return data_exists

    def find(self, exchange: str, base: str, quote: str, settle: str | None = None) -> Instrument | None:
        for i in self._lookup.values():
            if i.exchange == exchange and (
                (i.base == base and i.quote == quote) or (i.base == quote and i.quote == base)
            ):
                if settle is not None and i.settle is not None:
                    if i.settle == settle:
                        return i
                else:
                    return i
        return None

    def find_symbol(self, exchange: str, symbol: str) -> Instrument | None:
        for i in self._lookup.values():
            if (i.exchange == exchange) and (i.symbol == symbol):
                return i
        return None

    def find_instruments(self, exchange: str, quote: str | None = None) -> list[Instrument]:
        return [i for i in self._lookup.values() if i.exchange == exchange and (quote is None or i.quote == quote)]

    def _save_to_json(self, path, instruments: list[Instrument]):
        with open(path, "w") as f:
            json.dump(instruments, f, cls=_InstrumentEncoder, indent=4)
        logger.info(f"Saved {len(instruments)} to {path}")

    def find_aux_instrument_for(self, instrument: Instrument, base_currency: str) -> Instrument | None:
        """
        Tries to find aux instrument (for conversions to funded currency)
        for example:
            ETHBTC -> BTCUSDT for base_currency USDT
            EURGBP -> GBPUSD for base_currency USD
            ...
        """
        base_currency = base_currency.upper()
        if instrument.quote != base_currency:
            return self.find(instrument.exchange, instrument.quote, base_currency)
        return None

    def __getitem__(self, spath: str) -> list[Instrument]:
        res = []
        c = re.compile(spath)
        for k, v in self._lookup.items():
            if re.match(c, k):
                res.append(v)
        return res

    def refresh(self, query_exchanges: bool = False):
        for mn in dir(self):
            if mn.startswith("_update_"):
                getattr(self, mn)(self._path, query_exchanges)

    def _ccxt_update(
        self,
        path: str,
        file_name: str,
        exchange_to_ccxt_name: dict[str, str],
        keep_types: list[MarketType] | None = None,
        query_exchanges: bool = False,
    ):
        import ccxt as cx

        from qubx.utils.marketdata.ccxt import ccxt_symbol_to_instrument

        # - first we try to load packed data from QUBX resources
        instruments = {}
        _packed_data = _load_qubx_resources_as_json(f"instruments/symbols-{file_name}")
        if _packed_data:
            for i in _convert_instruments_metadata_to_qubx(_packed_data):
                instruments[i] = i

        if query_exchanges:
            # - replace defaults with data from CCXT
            for exch, ccxt_name in exchange_to_ccxt_name.items():
                exch = exch.upper()
                ccxt_name = ccxt_name.lower()
                ex: cx.Exchange = getattr(cx, ccxt_name)()
                mkts = ex.load_markets()
                for v in mkts.values():
                    if v["index"]:
                        continue
                    instr = ccxt_symbol_to_instrument(exch, v)
                    if not keep_types or instr.market_type in keep_types:
                        instruments[instr] = instr

        # - drop to file
        self._save_to_json(os.path.join(path, f"{file_name}.json"), list(instruments.values()))

    def _update_kraken(self, path: str, query_exchanges: bool = False):
        self._ccxt_update(path, "kraken.f", {"kraken.f": "krakenfutures"}, query_exchanges=query_exchanges)
        self._ccxt_update(path, "kraken", {"kraken": "kraken"}, query_exchanges=query_exchanges)

    def _update_hyperliquid(self, path: str, query_exchanges: bool = False):
        self._ccxt_update(
            path,
            "hyperliquid",
            {"hyperliquid": "hyperliquid"},
            keep_types=[MarketType.SPOT],
            query_exchanges=query_exchanges,
        )
        self._ccxt_update(
            path,
            "hyperliquid.f",
            {"hyperliquid.f": "hyperliquid"},
            keep_types=[MarketType.SWAP],
            query_exchanges=query_exchanges,
        )

    def _update_binance(self, path: str, query_exchanges: bool = False):
        self._ccxt_update(
            path,
            "binance",
            {"binance": "binance"},
            keep_types=[MarketType.SPOT, MarketType.MARGIN],
            query_exchanges=query_exchanges,
        )
        self._ccxt_update(path, "binance.um", {"binance.um": "binanceusdm"}, query_exchanges=query_exchanges)
        self._ccxt_update(path, "binance.cm", {"binance.cm": "binancecoinm"}, query_exchanges=query_exchanges)

    def _update_bitfinex(self, path: str, query_exchanges: bool = False):
        self._ccxt_update(
            path,
            "bitfinex.f",
            {"bitfinex.f": "bitfinex"},
            keep_types=[MarketType.SWAP],
            query_exchanges=query_exchanges,
        )

    def _update_dukas(self, path: str, query_exchanges: bool = False):
        self._save_to_json(os.path.join(path, "dukas.json"), SAMPLE_INSTRUMENTS)


# - TODO: need to find better way to extract actual data !!
_DEFAULT_FEES = """
[binance]
# SPOT (maker, taker)
vip0_usdt = 0.1000,0.1000
vip1_usdt = 0.0900,0.1000
vip2_usdt = 0.0800,0.1000
vip3_usdt = 0.0420,0.0600
vip4_usdt = 0.0420,0.0540
vip5_usdt = 0.0360,0.0480
vip6_usdt = 0.0300,0.0420
vip7_usdt = 0.0240,0.0360
vip8_usdt = 0.0180,0.0300
vip9_usdt = 0.0120,0.0240

# SPOT (maker, taker)
vip0_bnb = 0.0750,0.0750
vip1_bnb = 0.0675,0.0750
vip2_bnb = 0.0600,0.0750
vip3_bnb = 0.0315,0.0450
vip4_bnb = 0.0315,0.0405
vip5_bnb = 0.0270,0.0360
vip6_bnb = 0.0225,0.0315
vip7_bnb = 0.0180,0.0270
vip8_bnb = 0.0135,0.0225
vip9_bnb = 0.0090,0.0180

# UM futures (maker, taker)
[binance.um]
vip0_usdt = 0.0200,0.0500
vip1_usdt = 0.0160,0.0400
vip2_usdt = 0.0140,0.0350
vip3_usdt = 0.0120,0.0320
vip4_usdt = 0.0100,0.0300
vip5_usdt = 0.0080,0.0270
vip6_usdt = 0.0060,0.0250
vip7_usdt = 0.0040,0.0220
vip8_usdt = 0.0020,0.0200
vip9_usdt = 0.0000,0.0170

# CM futures (maker, taker)
[binance.cm]
vip0 = 0.0200,0.0500
vip1 = 0.0160,0.0400
vip2 = 0.0140,0.0350
vip3 = 0.0120,0.0320
vip4 = 0.0100,0.0300
vip5 = 0.0080,0.0270
vip6 = 0.0060,0.0250
vip7 = 0.0040,0.0220
vip8 = 0.0020,0.0200
vip9 = 0.0000,0.0170

[bitmex]
tierb_xbt=0.02,0.075
tierb_usdt=-0.015,0.075
tieri_xbt=0.01,0.05
tieri_usdt=-0.015,0.05
tiert_xbt=0.0,0.04
tiert_usdt=-0.015,0.04
tierm_xbt=0.0,0.035
tierm_usdt=-0.015,0.035
tiere_xbt=0.0,0.03
tiere_usdt=-0.015,0.03
tierx_xbt=0.0,0.025
tierx_usdt=-0.015,0.025
tierd_xbt=-0.003,0.024
tierd_usdt=-0.015,0.024
tierw_xbt=-0.005,0.023
tierw_usdt=-0.015,0.023
tierk_xbt=-0.008,0.022
tierk_usdt=-0.015,0.022
tiers_xbt=-0.01,0.0175
tiers_usdt=-0.015,0.02

[dukas]
regular=0.0035,0.0035
premium=0.0017,0.0017

[kraken]
K0=0.25,0.40
K10=0.20,0.35
K50=0.14,0.24
K100=0.12,0.22
K250=0.10,0.20
K500=0.08,0.18
M1=0.06,0.16
M2.5=0.04,0.14
M5=0.02,0.12
M10=0.0,0.10

[kraken.f]
K0=0.0200,0.0500
K100=0.0150,0.0400
M1=0.0125,0.0300
M5=0.0100,0.0250
M10=0.0075,0.0200
M20=0.0050,0.0150
M50=0.0025,0.0125
M100=0.0000,0.0100
"""


class FeesLookup:
    """
    Fees lookup
    """

    _lookup: dict[str, TransactionCostsCalculator]
    _path: str

    def __init__(self, path: str = makedirs(get_local_qubx_folder(), _DEF_FEES_FOLDER)) -> None:
        self._path = path
        if not self.load():
            self.refresh()
        self.load()

    def load(self) -> bool:
        self._lookup = {}
        data_exists = False
        parser = configparser.ConfigParser()
        # - load all avaliable configs
        for fs in glob.glob(self._path + "/*.ini"):
            parser.read(fs)
            data_exists = True

        for exch in parser.sections():
            for spec, info in parser[exch].items():
                try:
                    maker, taker = info.split(",")
                    self._lookup[f"{exch}_{spec}"] = (float(maker), float(taker))
                except (ValueError, TypeError) as e:
                    logger.warning(f'Wrong spec format for {exch}: "{info}". Should be spec=maker,taker. Error: {e}')

        return data_exists

    def __getitem__(self, spath: str) -> list[Instrument]:
        res = []
        c = re.compile(spath)
        for k, v in self._lookup.items():
            if re.match(c, k):
                res.append((k, v))
        return res

    def refresh(self):
        with open(os.path.join(self._path, "default.ini"), "w") as f:
            f.write(_DEFAULT_FEES)

    def find(self, exchange: str, spec: str | None) -> TransactionCostsCalculator | None:
        if spec is None:
            return ZERO_COSTS
        key = f"{exchange}_{spec}"
        vals = self._lookup.get(key)
        return TransactionCostsCalculator(key, *self._lookup.get(key)) if vals is not None else None

    def __repr__(self) -> str:
        s = "Name:\t\t\t(maker, taker)\n"
        for k, v in self._lookup.items():
            s += f"{k.ljust(25)}: {v}\n"
        return s


@dataclasses.dataclass(frozen=True)
class GlobalLookup:
    instruments: InstrumentsLookup
    fees: FeesLookup

    def find_fees(self, exchange: str, spec: str | None) -> TransactionCostsCalculator | None:
        return self.fees.find(exchange, spec)

    def find_aux_instrument_for(self, instrument: Instrument, base_currency: str) -> Instrument | None:
        return self.instruments.find_aux_instrument_for(instrument, base_currency)

    def find_instrument(self, exchange: str, base: str, quote: str) -> Instrument | None:
        return self.instruments.find(exchange, base, quote)

    def find_instruments(self, exchange: str, quote: str | None = None) -> list[Instrument]:
        return self.instruments.find_instruments(exchange, quote)

    def find_symbol(self, exchange: str, symbol: str) -> Instrument | None:
        return self.instruments.find_symbol(exchange, symbol)


def _load_qubx_resources_as_json(path: Path | str) -> list[dict]:
    """
    Reads a JSON file from resource module
    """
    import importlib.resources
    import json

    if isinstance(path, str):
        path = Path(path)

    if path.suffix != ".json":
        path = path.with_suffix(path.suffix + ".json")

    data = []
    try:
        res_path = importlib.resources.files("qubx.resources") / path
        with res_path.open() as f:
            data = json.load(f)
    except Exception as e:
        logger.warning(f"Can't load resource file from {path} - {str(e)}")

    return data


def _convert_instruments_metadata_to_qubx(data: list[dict]):
    """
    Converting tardis symbols meta-data to Qubx instruments
    """
    _excs = {
        "binance": "BINANCE",
        "binance-delivery": "BINANCE.CM",
        "binance-futures": "BINANCE.UM",
        "kraken": "KRAKEN",
        "cryptofacilities": "KRAKEN.F",
        "bitfinex": "BITFINEX",
        "bitfinex-derivatives": "BITFINEX.F",
        "hyperliquid": "HYPERLIQUID",
    }
    r = []
    for s in data:
        match s["type"]:
            case "perpetual":
                _type = MarketType.SWAP
            case "spot":
                _type = MarketType.SPOT
            case "future":
                _type = MarketType.FUTURE
            case _:
                raise ValueError(f" -> Unknown type {s['type']}")
        r.append(
            Instrument(
                s["baseCurrency"] + s["quoteCurrency"],
                AssetType.CRYPTO,
                _type,
                _excs.get(s["exchange"], s["exchange"].upper()),
                s["baseCurrency"],
                s["quoteCurrency"],
                s["quoteCurrency"],
                s["id"],
                tick_size=s["priceIncrement"],
                lot_size=s["minTradeAmount"],
                min_size=s["amountIncrement"],
                min_notional=0,  # we don't have this info from tardis
                contract_size=s.get("contractMultiplier", 1.0),
                onboard_date=s.get("availableSince", None),
                delivery_date=s.get("availableTo", None),
            )
        )
    return r


# - global lookup helper
lookup = GlobalLookup(InstrumentsLookup(), FeesLookup())



""" "
Here stuff we want to have in every Jupyter notebook after calling %qubx magic
"""

import qubx
from qubx import runtime_env
from qubx.utils.misc import add_project_to_system_path, logo


def np_fmt_short():
    # default np output is 75 columns so extend it a bit and suppress scientific fmt for small floats
    np.set_printoptions(linewidth=240, suppress=True)


def np_fmt_reset():
    # reset default np printing options
    np.set_printoptions(
        edgeitems=3,
        infstr="inf",
        linewidth=75,
        nanstr="nan",
        precision=8,
        suppress=False,
        threshold=1000,
        formatter=None,
    )


if runtime_env() in ["notebook", "shell"]:
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # -- all imports below will appear in notebook after calling %%qubx magic ---
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    # - - - - Common stuff - - - -
    from datetime import time, timedelta  # noqa: F401

    import numpy as np  # noqa: F401
    import pandas as pd  # noqa: F401

    # - - - - Charting stuff - - - -
    from matplotlib import pyplot as plt  # noqa: F401
    from tqdm.auto import tqdm  # noqa: F401

    # - - - - TA stuff and indicators - - - -
    import qubx.pandaz.ta as pta  # noqa: F401
    import qubx.ta.indicators as ta  # noqa: F401
    from qubx.backtester.optimization import variate  # noqa: F401

    # - - - - Simulator stuff - - - -
    from qubx.backtester.simulator import simulate  # noqa: F401

    # - - - - Portfolio analysis - - - -
    from qubx.core.metrics import (  # noqa: F401
        chart_signals,
        drop_symbols,
        get_symbol_pnls,
        pick_symbols,
        pnl,
        portfolio_metrics,
        tearsheet,
    )
    from qubx.data.helpers import loader  # noqa: F401

    # - - - - Data reading - - - -
    from qubx.data.readers import (  # noqa: F401
        AsOhlcvSeries,
        AsPandasFrame,
        AsQuotes,
        AsTimestampedRecords,
        CsvStorageDataReader,
        MultiQdbConnector,
        QuestDBConnector,
        RestoreTicksFromOHLC,
    )

    # - - - - Utils - - - -
    from qubx.pandaz.utils import (  # noqa: F401
        continuous_periods,
        drop_duplicated_indexes,
        generate_equal_date_ranges,
        ohlc_resample,
        retain_columns_and_join,
        rolling_forward_test_split,
        scols,
        srows,
    )
    from qubx.utils.charting.lookinglass import LookingGlass  # noqa: F401
    from qubx.utils.charting.mpl_helpers import fig, ohlc_plot, plot_trends, sbp, subplot  # noqa: F401
    from qubx.utils.misc import this_project_root  # noqa: F401

    # - setup short numpy output format
    np_fmt_short()

    # - add project home to system path
    add_project_to_system_path()

    # show logo first time
    if not hasattr(qubx.QubxMagics, "__already_initialized__"):
        setattr(qubx.QubxMagics, "__already_initialized__", True)
        logo()



from qubx.core.basics import DataType, Instrument, TargetPosition
from qubx.core.helpers import CachedMarketDataHolder
from qubx.core.interfaces import (
    IAccountProcessor,
    IBroker,
    IDataProvider,
    IPositionGathering,
    IStrategy,
    IStrategyContext,
    ISubscriptionManager,
    ITimeProvider,
    ITradingManager,
    IUniverseManager,
    RemovalPolicy,
)
from qubx.core.loggers import StrategyLogging


class UniverseManager(IUniverseManager):
    _context: IStrategyContext
    _strategy: IStrategy
    _broker: IDataProvider
    _trading_service: IBroker
    _cache: CachedMarketDataHolder
    _logging: StrategyLogging
    _subscription_manager: ISubscriptionManager
    _trading_manager: ITradingManager
    _time_provider: ITimeProvider
    _account: IAccountProcessor
    _position_gathering: IPositionGathering
    _removal_queue: dict[Instrument, tuple[RemovalPolicy, bool]]

    def __init__(
        self,
        context: IStrategyContext,
        strategy: IStrategy,
        broker: IDataProvider,
        trading_service: IBroker,
        cache: CachedMarketDataHolder,
        logging: StrategyLogging,
        subscription_manager: ISubscriptionManager,
        trading_manager: ITradingManager,
        time_provider: ITimeProvider,
        account: IAccountProcessor,
        position_gathering: IPositionGathering,
    ):
        self._context = context
        self._strategy = strategy
        self._broker = broker
        self._trading_service = trading_service
        self._cache = cache
        self._logging = logging
        self._subscription_manager = subscription_manager
        self._trading_manager = trading_manager
        self._time_provider = time_provider
        self._account = account
        self._position_gathering = position_gathering
        self._instruments = []
        self._removal_queue = {}

    def _has_position(self, instrument: Instrument) -> bool:
        return (
            instrument in self._account.positions
            and abs(self._account.positions[instrument].quantity) > instrument.min_size
        )

    def set_universe(
        self,
        instruments: list[Instrument],
        skip_callback: bool = False,
        if_has_position_then: RemovalPolicy = "close",
    ) -> None:
        assert if_has_position_then in (
            "close",
            "wait_for_close",
            "wait_for_change",
        ), "Invalid if_has_position_then policy"

        new_set = set(instruments)
        prev_set = set(self._instruments)

        # - determine instruments to remove depending on if_has_position_then policy
        may_be_removed = list(prev_set - new_set)

        # - split instruments into removable and keepable
        to_remove, to_keep = self._get_what_can_be_removed_or_kept(may_be_removed, skip_callback, if_has_position_then)

        to_add = list(new_set - prev_set)
        self.__do_add_instruments(to_add)
        self.__do_remove_instruments(to_remove)

        # - cleanup removal queue
        self.__cleanup_removal_queue(instruments)

        if not skip_callback and (to_add or to_remove):
            self._strategy.on_universe_change(self._context, to_add, to_remove)

        self._subscription_manager.commit()  # apply pending changes

        # set new instruments
        self._instruments.clear()
        self._instruments.extend(instruments)
        self._instruments.extend(to_keep)

    def _get_what_can_be_removed_or_kept(
        self, may_be_removed: list[Instrument], skip_callback: bool, if_has_position_then: RemovalPolicy
    ) -> tuple[list[Instrument], list[Instrument]]:
        immediately_close = if_has_position_then == "close"
        to_remove, to_keep = [], []
        for instr in may_be_removed:
            if immediately_close:
                to_remove.append(instr)
            else:
                if self._has_position(instr):
                    self._removal_queue[instr] = (if_has_position_then, skip_callback)
                    to_keep.append(instr)
        return to_remove, to_keep

    def __cleanup_removal_queue(self, instruments: list[Instrument]):
        for instr in instruments:
            # - if it's still in the removal queue, remove it
            if instr in self._removal_queue:
                self._removal_queue.pop(instr)

    def add_instruments(self, instruments: list[Instrument]):
        self.__do_add_instruments(instruments)
        self.__cleanup_removal_queue(instruments)
        self._strategy.on_universe_change(self._context, instruments, [])
        self._subscription_manager.commit()
        self._instruments.extend(instruments)

    def remove_instruments(
        self,
        instruments: list[Instrument],
        if_has_position_then: RemovalPolicy = "close",
    ):
        assert if_has_position_then in (
            "close",
            "wait_for_close",
            "wait_for_change",
        ), "Invalid if_has_position_then policy"

        # - split instruments into removable and keepable
        to_remove, to_keep = self._get_what_can_be_removed_or_kept(instruments, False, if_has_position_then)

        # - remove ones that can be removed immediately
        self.__do_remove_instruments(to_remove)
        self._strategy.on_universe_change(self._context, [], to_remove)
        self._subscription_manager.commit()

        # - update instruments list
        self._instruments = list(set(self._instruments) - set(to_remove))
        self._instruments.extend(to_keep)

    @property
    def instruments(self) -> list[Instrument]:
        return self._instruments

    def __do_remove_instruments(self, instruments: list[Instrument]):
        """
        Remove symbols from universe. Steps:
        - [v] cancel all open orders
        - [v] close all open positions
        - [v] unsubscribe from market data
        - [v] remove from data cache

        We are still keeping the symbols in the positions dictionary.
        """
        if not instruments:
            return

        # - preprocess instruments and cancel all open orders
        for instr in instruments:
            # - remove instrument from the removal queue if it's there
            self._removal_queue.pop(instr, None)

            # - cancel all open orders
            self._trading_manager.cancel_orders(instr)

        # - close all open positions
        exit_targets = [
            TargetPosition.zero(self._context, instr.signal(0, group="Universe", comment="Universe change"))
            for instr in instruments
            if self._has_position(instr)
        ]
        self._position_gathering.alter_positions(self._context, exit_targets)

        # - if still open positions close them manually
        for instr in instruments:
            pos = self._account.positions.get(instr)
            if pos and abs(pos.quantity) > instr.min_size:
                self._trading_manager.trade(instr, -pos.quantity)

        # - unsubscribe from market data
        for instr in instruments:
            self._subscription_manager.unsubscribe(DataType.ALL, instr)

        # - remove from data cache
        for instr in instruments:
            self._cache.remove(instr)

    def __do_add_instruments(self, instruments: list[Instrument]) -> None:
        # - create positions for instruments
        self._create_and_update_positions(instruments)

        # - initialize ohlcv for new instruments
        for instr in instruments:
            self._cache.init_ohlcv(instr)

        # - subscribe to market data
        self._subscription_manager.subscribe(
            (
                DataType.ALL
                if self._subscription_manager.auto_subscribe
                else self._subscription_manager.get_base_subscription()
            ),
            instruments,
        )

        # - reinitialize strategy loggers
        self._logging.initialize(self._time_provider.time(), self._account.positions, self._account.get_balances())

    def _create_and_update_positions(self, instruments: list[Instrument]):
        for instrument in instruments:
            _ = self._account.get_position(instrument)

            # - check if we need any aux instrument for calculating pnl ?
            # TODO: test edge cases for aux symbols (UniverseManager)
            # aux = lookup.find_aux_instrument_for(instrument, self._account.get_base_currency())
            # if aux is not None:
            #     instrument._aux_instrument = aux
            #     instruments.append(aux)
            #     _ = self._trading_service.get_position(aux)

    def on_alter_position(self, instrument: Instrument) -> None:
        """
        Called when the position of an instrument changes.
        It can be used for postponed unsubscribed events
        """
        # - check if need to remove instrument from the universe
        if instrument in self._removal_queue:
            _, skip_callback = self._removal_queue[instrument]

            # - if no position, remove instrument from the universe
            if not self._has_position(instrument):
                self.__do_remove_instruments([instrument])

                if not skip_callback:
                    self._strategy.on_universe_change(self._context, [], [instrument])

                # - commit changes and remove instrument from the universe
                self._subscription_manager.commit()
                self._instruments.remove(instrument)

    def is_trading_allowed(self, instrument: Instrument) -> bool:
        if instrument in self._removal_queue:
            policy, skip_callback = self._removal_queue[instrument]

            if policy == "wait_for_change":
                self.__do_remove_instruments([instrument])

                if not skip_callback:
                    self._strategy.on_universe_change(self._context, [], [instrument])

                # - commit changes and remove instrument from the universe
                self._subscription_manager.commit()
                self._instruments.remove(instrument)
                return False

        return True



import ast
import getpass
import os
import shutil
import sys
from collections import namedtuple
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Generator, List, Tuple

import yaml
from git import Repo

from qubx import logger
from qubx.utils.misc import cyan, green, magenta, makedirs, red, white, yellow
from qubx.utils.runner.configs import ExchangeConfig, LoggingConfig, StrategyConfig, load_strategy_config_from_yaml

from .misc import (
    StrategyInfo,
    find_git_root,
    find_pyproject_root,
    scan_strategies_in_directory,
)

Import = namedtuple("Import", ["module", "name", "alias"])
DEFAULT_CFG_NAME = "config.yml"


@dataclass
class ReleaseInfo:
    tag: str
    commit: str
    user: str
    time: datetime
    commited_files: list[str]


def get_imports(path, what_to_look=["xincubator"]) -> Generator[Import, None, None]:
    """
    Get imports from the given file.
    """
    with open(path) as fh:
        root = ast.parse(fh.read(), path)

    for node in ast.iter_child_nodes(root):
        if isinstance(node, ast.Import):
            module = []
        elif isinstance(node, ast.ImportFrom):
            module = node.module.split(".") if node.module else []
        else:
            continue
        for n in node.names:
            if module and what_to_look and module[0] in what_to_look:
                yield Import(module, n.name.split("."), n.asname)


def ls_strats(path: str) -> None:
    """
    List all strategies in the given directory.
    """
    strats = scan_strategies_in_directory(path)
    for si in strats:
        strs = ""
        descr = (": " + green(si.docstring.replace("\n", " ").strip('" '))) if si.docstring else ""
        _p_str = ""
        if si.parameters:
            _max_l = max(map(len, si.parameters.keys())) + 2
            for k, c in si.parameters.items():
                _p_str += f"\t{red(':')}  {cyan(k.ljust(_max_l))}: {yellow(str(c))}\n"
            strs += f"\t{red('.--(')} {white(si.name)} {descr} \n{_p_str}\n"

        rst = f""" - {magenta(si.path)} -
{strs}"""
        print(rst)


def _find_strategies(strats: list[StrategyInfo], strategy_name: str) -> list[StrategyInfo]:
    """
    Filter strategies by name. If there are multiple strategies with the same name, the first one is returned.

    Args:
        strats: dict[str, list[tuple[str, str, dict]]] - strategies to filter
        strategy_name: str - strategy name to filter

    Returns:
        list[tuple[str, StrategyInfo]] - filtered strategies, each tuple contains (path, strategy info)
    """
    _filt_strats = []

    for si in strats:
        _s_name = si.name
        if _s_name.lower() == strategy_name.lower():
            _filt_strats.append(si)

    return _filt_strats


def is_config_file(path: str) -> bool:
    """
    Check if the given path is a YAML config file.

    Args:
        path: Path to check

    Returns:
        True if the path is a YAML config file, False otherwise
    """
    config_path = Path(path)
    return config_path.exists() and config_path.is_file() and (config_path.suffix.lower() in [".yml", ".yaml"])


def find_strategy_by_name(directory: str, strategy_name: str) -> StrategyInfo:
    """
    Find a strategy by name in the given directory.

    Args:
        directory: Directory to scan for strategies
        strategy_name: Strategy name to find

    Returns:
        StrategyInfo for the found strategy

    Raises:
        ValueError: If no strategy is found or multiple strategies are found
    """
    strats = scan_strategies_in_directory(directory)
    matching_strats = _find_strategies(strats, strategy_name)

    if not matching_strats:
        raise ValueError(f"Strategy {strategy_name} not found in {directory}")

    if len(matching_strats) > 1:
        strat_paths = [si.path for si in matching_strats]
        strat_paths_str = "\n".join([f"\t - {path}" for path in strat_paths])
        raise ValueError(f"Multiple strategies found for {strategy_name}:\n{strat_paths_str}")

    return matching_strats[0]


def generate_default_config(
    stg_info: StrategyInfo, exchange: str, connector: str, instruments: List[str]
) -> StrategyConfig:
    """
    Generate a default configuration for the strategy.

    Args:
        stg_info: Strategy information
        exchange: Default exchange
        connector: Default connector
        instruments: Default instruments

    Returns:
        Default StrategyConfig
    """
    # Create exchange config
    exchange_config = ExchangeConfig(connector=connector, universe=instruments)

    # Create logging config
    logging_config = LoggingConfig(logger="CsvFileLogsWriter", position_interval="10Sec", portfolio_interval="5Min")

    # Generate the import path for the strategy
    strategy_path = stg_info.path

    # Get the pyproject root directory
    pyproject_root = find_pyproject_root(strategy_path)
    package_name = os.path.basename(pyproject_root)

    # Get the relative path from the pyproject root to the strategy file
    rel_path = os.path.relpath(strategy_path, pyproject_root)

    # Convert file path to module path
    # Remove .py extension and replace path separators with dots
    module_path = os.path.splitext(rel_path)[0].replace(os.path.sep, ".")

    # Construct the full import path
    import_path = f"{module_path}.{stg_info.name}"

    # If the module path starts with the package name, we don't need to repeat it
    if import_path.startswith(f"{package_name}."):
        import_path = import_path
    else:
        import_path = f"{package_name}.{import_path}"

    # Create strategy config
    strategy_config = StrategyConfig(
        strategy=import_path,
        parameters=stg_info.parameters,
        exchanges={exchange: exchange_config},
        logging=logging_config,
    )

    return strategy_config


def load_strategy_from_config(config_path: Path, directory: str) -> Tuple[StrategyInfo, StrategyConfig]:
    """
    Load strategy information from a config file.

    Args:
        config_path: Path to the config file
        directory: Directory to scan for strategies

    Returns:
        Tuple of (StrategyInfo, StrategyConfig)
    """
    try:
        # Load the config using the utility function
        strategy_config = load_strategy_config_from_yaml(config_path)

        # Extract strategy name from config
        strategy_path = strategy_config.strategy
        if not strategy_path:
            raise ValueError("Strategy path not found in config file")

        # Get the strategy class name (last part of the path)
        strategy_name = strategy_path.split(".")[-1]

        # Find the strategy
        strategy_info = find_strategy_by_name(directory, strategy_name)

        return strategy_info, strategy_config
    except Exception as e:
        logger.error(f"Error loading strategy from config file: {e}")
        raise


def release_strategy(
    directory: str,
    strategy_name: str,
    tag: str | None,
    message: str | None,
    commit: bool,
    output_dir: str,
    default_exchange: str = "BINANCE.UM",
    default_connector: str = "ccxt",
    default_instruments: List[str] = ["BTCUSDT"],
) -> None:
    """
    Release strategy to zip file from given directory or config file

    Args:
        directory: str - directory to scan for strategies
        strategy_name: str - strategy name to release or path to config file
        tag: str - additional tag for this release
        message: str - release message
        commit: bool - commit changes and create tag in repo
        output_dir: str - output directory to put zip file
        default_exchange: str - default exchange to use in the generated config
        default_connector: str - default connector to use in the generated config
        default_instruments: List[str] - default instruments to use in the generated config
    """
    try:
        # Determine if strategy_name is a config file or a strategy name
        if is_config_file(strategy_name):
            # Load strategy from config file
            logger.info(f"Loading strategy from config file: {strategy_name}")
            stg_info, strategy_config = load_strategy_from_config(Path(strategy_name), directory)
        else:
            # Find strategy by name
            logger.info(f"Finding strategy by name: {strategy_name}")
            stg_info = find_strategy_by_name(directory, strategy_name)

            # Generate default config
            strategy_config = generate_default_config(
                stg_info, default_exchange, default_connector, default_instruments
            )

        # Get repo root path
        stg_path = stg_info.path
        repo_path = find_git_root(stg_path)
        pyproject_root = find_pyproject_root(stg_path)

        # Process repo
        _skip_tag, _skip_commit = not commit, not commit

        _git_info = process_git_repo(
            repo_path=repo_path,
            subdir=pyproject_root,
            strategy_name_id=stg_info.name,
            tag_sfx=tag,
            message=message,
            skip_tag=_skip_tag,
            skip_commit=_skip_commit,
        )

        # Create zipped pack
        create_released_pack(
            stg_info=stg_info,
            git_info=_git_info,
            pyproject_root=pyproject_root,
            output_dir=output_dir,
            strategy_config=strategy_config,
        )
    except ValueError as e:
        logger.error(f"<r>{str(e)}</r>")
    except Exception as e:
        logger.error(f"<r>Error releasing strategy: {e}</r>")


def create_released_pack(
    stg_info: StrategyInfo,
    git_info: ReleaseInfo,
    pyproject_root: str,
    output_dir: str,
    strategy_config: StrategyConfig,
):
    """
    Create a release package for a strategy.

    Args:
        stg_info: Strategy information
        git_info: Git release information
        pyproject_root: Path to the pyproject.toml root directory
        output_dir: Output directory for the release package
        strategy_config: Strategy configuration
    """
    logger.info(f"Creating release pack for {git_info.tag} ...")

    # Setup directory structure
    stg_name = stg_info.name
    release_dir = makedirs(output_dir, git_info.tag)

    # Copy strategy file
    _copy_strategy_file(stg_info.path, pyproject_root, release_dir)

    # Copy dependencies
    _copy_dependencies(stg_info.path, pyproject_root, release_dir)

    # Save configuration
    _save_strategy_config(stg_name, strategy_config, release_dir)

    # Create metadata
    _create_metadata(stg_name, git_info, release_dir)

    # Handle project files
    _handle_project_files(pyproject_root, release_dir)

    # Create zip archive
    _create_zip_archive(output_dir, release_dir, git_info.tag)

    logger.info(f"Created release pack: {os.path.join(output_dir, git_info.tag)}.zip")


def _save_strategy_config(stg_name: str, strategy_config: StrategyConfig, release_dir: str) -> None:
    """Save the strategy configuration to the release directory."""
    config_path = os.path.join(release_dir, "config.yml")
    with open(config_path, "wt") as fs:
        # Convert to dict and save directly (not under 'config' key)
        config_dict = strategy_config.model_dump()
        yaml.safe_dump(config_dict, fs, sort_keys=False, indent=2)
    logger.debug(f"Saved strategy config to {config_path}")


def _copy_strategy_file(strategy_path: str, pyproject_root: str, release_dir: str) -> None:
    """Copy the strategy file to the release directory."""
    src_dir = os.path.basename(pyproject_root)
    rel_path = os.path.relpath(strategy_path, pyproject_root)
    dest_file_path = os.path.join(release_dir, src_dir, rel_path)

    # Ensure the destination directory exists
    os.makedirs(os.path.dirname(dest_file_path), exist_ok=True)

    # Copy strategy file
    logger.debug(f"Copying strategy file from {strategy_path} to {dest_file_path}")
    shutil.copy2(strategy_path, dest_file_path)


def _copy_dependencies(strategy_path: str, pyproject_root: str, release_dir: str) -> None:
    """Copy all dependencies required by the strategy."""
    src_dir = os.path.basename(pyproject_root)
    imports = _get_imports(strategy_path, pyproject_root, [src_dir])

    for imp in imports:
        # Construct source path
        src_i = os.path.join(pyproject_root, *[s for s in imp.module if s != src_dir]) + ".py"
        if not os.path.exists(src_i):
            src_i = os.path.join(pyproject_root, *[s for s in imp.module if s != src_dir], "__init__.py")
            if not os.path.exists(src_i):
                logger.warning(f"Import file not found: {src_i}")
                continue

        # Get the relative path from pyproject_root
        rel_import_path = os.path.relpath(src_i, pyproject_root)
        dest_import_path = os.path.join(release_dir, src_dir, rel_import_path)

        # Ensure the destination directory exists
        os.makedirs(os.path.dirname(dest_import_path), exist_ok=True)

        # Copy the import file
        logger.debug(f"Copying import from {src_i} to {dest_import_path}")
        shutil.copy2(src_i, dest_import_path)


def _create_metadata(stg_name: str, git_info: ReleaseInfo, release_dir: str) -> None:
    """Create metadata files for the release."""
    # Create meta info file
    with open(os.path.join(release_dir, f"{stg_name}.info"), "wt") as fs:
        yaml.safe_dump(
            {
                "tag": git_info.tag,
                "date": git_info.time.isoformat(),
                "author": git_info.user,
            },
            fs,
            sort_keys=False,
        )

    # Create a README.md file
    with open(os.path.join(release_dir, "README.md"), "wt") as fs:
        fs.write(f"# {stg_name}\n\n")
        fs.write("## Git Info\n\n")
        fs.write(f"Tag: {git_info.tag}\n")
        fs.write(f"Date: {git_info.time.isoformat()}\n")
        fs.write(f"Author: {git_info.user}\n")
        fs.write(f"Commit: {git_info.commit}\n")


def _modify_pyproject_toml(pyproject_path: str, package_name: str) -> None:
    """
    Modify the pyproject.toml file to include the project package as a dependency.

    Args:
        pyproject_path: Path to the pyproject.toml file
        package_name: Name of the package to add as a dependency
    """
    try:
        import toml

        # Read the existing pyproject.toml
        with open(pyproject_path, "r") as f:
            pyproject_data = toml.load(f)

        # Add the package as a local dependency
        if "tool" in pyproject_data and "poetry" in pyproject_data["tool"]:
            # Ensure dependencies section exists
            if "dependencies" not in pyproject_data["tool"]["poetry"]:
                pyproject_data["tool"]["poetry"]["dependencies"] = {}

            # Add Python as a dependency if not already present
            if "python" not in pyproject_data["tool"]["poetry"]["dependencies"]:
                python_version = f"{sys.version_info.major}.{sys.version_info.minor}"
                pyproject_data["tool"]["poetry"]["dependencies"]["python"] = f"^{python_version}"

            # Replace the packages section with the new one
            pyproject_data["tool"]["poetry"]["packages"] = [{"include": package_name}]

            # Write the updated pyproject.toml
            with open(pyproject_path, "w") as f:
                toml.dump(pyproject_data, f)

            logger.debug(f"Updated pyproject.toml to include {package_name} as a local dependency")
    except Exception as e:
        logger.warning(f"Failed to update pyproject.toml: {e}")


def _generate_poetry_lock(release_dir: str) -> None:
    """
    Generate a poetry.lock file without creating a virtual environment.

    Args:
        release_dir: Directory where the poetry.lock file should be generated
    """
    import subprocess

    try:
        # Configure Poetry settings for lock generation
        logger.debug("Configuring Poetry settings for lock generation without venv creation")

        # Set virtualenvs.create=false to prevent venv creation during lock generation
        subprocess.run(
            ["poetry", "config", "virtualenvs.create", "false", "--local"],
            cwd=release_dir,
            check=True,
            capture_output=True,
            text=True,
        )

        # Set virtualenvs.in-project=true for when the venv is eventually created during deployment
        subprocess.run(
            ["poetry", "config", "virtualenvs.in-project", "true", "--local"],
            cwd=release_dir,
            check=True,
            capture_output=True,
            text=True,
        )

        # Check if we're already in a Poetry shell
        in_poetry_env = "POETRY_ACTIVE" in os.environ or "VIRTUAL_ENV" in os.environ

        logger.info("Generating poetry.lock file without creating virtual environment...")

        # If we're in a Poetry shell, we need to be more explicit about avoiding environment creation
        # Add --no-interaction to prevent any prompts
        lock_cmd = ["poetry", "lock", "--no-update", "--no-interaction"]
        if in_poetry_env:
            # Force Poetry to use a clean environment even if we're in an active one
            logger.debug("Detected active Poetry environment. Using a clean environment for lock generation.")
            env = os.environ.copy()
            # Temporarily unset Poetry environment variables to avoid interference
            for var in ["POETRY_ACTIVE", "VIRTUAL_ENV"]:
                if var in env:
                    del env[var]

            subprocess.run(lock_cmd, cwd=release_dir, check=True, capture_output=True, text=True, env=env)
        else:
            # Normal case - not in a Poetry shell
            subprocess.run(lock_cmd, cwd=release_dir, check=True, capture_output=True, text=True)

        # After lock generation, reset the virtualenvs.create setting to true for deployment
        # This ensures that when the package is deployed, the venv can be created
        subprocess.run(
            ["poetry", "config", "virtualenvs.create", "true", "--local"],
            cwd=release_dir,
            check=True,
            capture_output=True,
            text=True,
        )
    except Exception as e:
        logger.warning(f"Failed to generate poetry.lock: {e}")
        raise e


def _handle_project_files(pyproject_root: str, release_dir: str) -> None:
    """Handle project files like pyproject.toml and generate lock file."""
    # Copy pyproject.toml if it exists
    pyproject_src = os.path.join(pyproject_root, "pyproject.toml")
    if not os.path.exists(pyproject_src):
        raise FileNotFoundError(f"pyproject.toml not found in {pyproject_root}")

    pyproject_dest = os.path.join(release_dir, "pyproject.toml")
    logger.debug(f"Copying pyproject.toml from {pyproject_src} to {pyproject_dest}")
    shutil.copy2(pyproject_src, pyproject_dest)

    # Get the basename of the pyproject_root as the package name
    package_name = os.path.basename(pyproject_root)

    # Modify the pyproject.toml to include the project package
    _modify_pyproject_toml(pyproject_dest, package_name)

    # Generate the poetry.lock file
    _generate_poetry_lock(release_dir)


def _create_zip_archive(output_dir: str, release_dir: str, tag: str) -> None:
    """Create a zip archive of the release package."""
    logger.debug("Creating zip file...")
    file_path = os.path.join(output_dir, tag)
    shutil.make_archive(file_path, "zip", release_dir)
    shutil.rmtree(release_dir)


def _get_imports(file_name: str, current_directory: str, what_to_look: list[str]) -> list[Import]:
    imports = list(get_imports(file_name, what_to_look))
    current_dirname = os.path.basename(current_directory)
    for i in imports:
        try:
            f1 = os.path.join(*[current_directory, *[s for s in i.module if s != current_dirname]]) + ".py"
            if not os.path.exists(f1):
                f1 = (
                    os.path.join(*[current_directory, *[s for s in i.module if s != current_dirname], "__init__"])
                    + ".py"
                )
            imports.extend(_get_imports(f1, current_directory, what_to_look))
        except Exception:
            pass
    return imports


def generate_tag(strategy_name_id: str, tag_sfx: str | None) -> str:
    _tn = datetime.now()
    tag_s = f".{tag_sfx}" if tag_sfx else ""
    _strategy_name_id = strategy_name_id.replace(",", "_")
    tag = f"R_{_strategy_name_id}_{_tn.strftime('%Y%m%d%H%M%S')}{tag_s}"
    return tag


def make_tag_in_repo(repo: Repo, strategy_name_id: str, user: str, tag: str) -> str:
    """
    Create an annotated tag in the repository and push it to the remote.

    Args:
        repo: Git repository object
        strategy_name_id: Strategy name identifier
        user: Username
        tag: Tag name

    Returns:
        str: The tag name
    """
    repo.config_writer().set_value("push", "followTags", "true").release()

    _tn = datetime.now()
    _ = repo.create_tag(
        tag,
        message=f"Release of '{strategy_name_id}' at {_tn.strftime('%Y-%b-%d %H:%M:%S')} by {user}",
    )
    repo.remote("origin").push(f"refs/tags/{tag}")
    return tag


def process_git_repo(
    repo_path: str,
    subdir: str,
    strategy_name_id: str,
    tag_sfx: str | None = None,
    message: str | None = None,
    skip_tag: bool = False,
    skip_commit: bool = False,
) -> ReleaseInfo:
    """
    Process the git repository to get the release information and commit the changes.

    Args:
        repo_path: str - path to the git repository
        subdir: str - subdirectory to scan for changes
        strategy_name_id: str - strategy name id
        tag_sfx: str | None - tag suffix
        message: str | None - message
        skip_tag: bool - skip tag
        skip_commit: bool - skip commit

    Returns:
        ReleaseInfo - release information
    """
    repo = Repo(repo_path)
    _to_add: list[str] = []

    # Convert subdir to absolute path for comparison
    abs_subdir = os.path.abspath(subdir)

    if diffs := repo.index.diff(None):
        logger.info(" - Modified files -")
        for d in diffs:
            # Convert the file path to absolute path for comparison
            file_path = str(d.a_path) if d.a_path is not None else ""
            abs_file_path = os.path.abspath(os.path.join(repo_path, file_path))
            # Check if the file is under subdir
            if abs_file_path.startswith(abs_subdir):
                logger.info(f"  - {red(file_path)}")
                _to_add.append(file_path)

    if untr := repo.untracked_files:
        logger.info(" - Untracked files -")
        for d in untr:
            # Convert the file path to absolute path for comparison
            file_path = str(d)
            abs_file_path = os.path.abspath(os.path.join(repo_path, file_path))
            # Check if the file is under subdir
            if abs_file_path.startswith(abs_subdir):
                logger.info(f"  - {red(file_path)}")
                _to_add.append(file_path)

    user = getpass.getuser()
    tag = None
    commit_sha = None
    _tn = datetime.now()
    if _to_add:
        if not skip_commit:
            # - add changed files
            logger.info(green(f"Commiting changes for {len(_to_add)} files ... "))
            try:
                repo.index.add(_to_add)
                cmt = repo.index.commit(
                    f"Changes before release of '{strategy_name_id}' at {_tn.strftime('%Y-%b-%d %H:%M:%S')} by {user}."
                    f"{'' + message if message else ''}"
                )
                commit_sha = cmt.hexsha
                _ilist = repo.remotes[0].push()
                for i in _ilist:
                    logger.debug(f"\t{yellow(i.summary)}")
            except Exception as e:
                logger.error(f"Error committing changes: {e}")
                raise e

    if commit_sha is None:
        # get latest commit sha
        commit_sha = repo.head.commit.hexsha
        if _to_add and skip_commit:
            logger.warning("<r> >> Commiting changes is skipped due to --skip-commit option</r>")

    # - add annotated tag
    tag = generate_tag(strategy_name_id, tag_sfx)
    if not skip_tag:
        tag = make_tag_in_repo(repo, strategy_name_id, user, tag)
    else:
        logger.warning("<r> >> Creating git tag is skipped due to --skip-tag option</r>")

    return ReleaseInfo(tag=tag, commit=commit_sha, user=user, time=_tn, commited_files=_to_add)



"""
Formatters for exporting trading data.

This module provides interfaces and implementations for formatting trading data
before it is exported to external systems.
"""

from qubx.exporters.formatters.base import DefaultFormatter, IExportFormatter
from qubx.exporters.formatters.slack import SlackMessageFormatter

__all__ = ["IExportFormatter", "DefaultFormatter", "SlackMessageFormatter"]



from qubx import logger
from qubx.backtester.ome import OrdersManagementEngine
from qubx.core.account import BasicAccountProcessor
from qubx.core.basics import (
    ZERO_COSTS,
    CtrlChannel,
    Instrument,
    Order,
    Position,
    Timestamped,
    TransactionCostsCalculator,
    dt_64,
)
from qubx.core.interfaces import ITimeProvider
from qubx.core.series import Bar, OrderBook, Quote, Trade


class SimulatedAccountProcessor(BasicAccountProcessor):
    ome: dict[Instrument, OrdersManagementEngine]
    order_to_instrument: dict[str, Instrument]

    _channel: CtrlChannel
    _fill_stop_order_at_price: bool
    _half_tick_size: dict[Instrument, float]

    def __init__(
        self,
        account_id: str,
        channel: CtrlChannel,
        base_currency: str,
        initial_capital: float,
        time_provider: ITimeProvider,
        tcc: TransactionCostsCalculator = ZERO_COSTS,
        accurate_stop_orders_execution: bool = False,
    ) -> None:
        super().__init__(
            account_id=account_id,
            time_provider=time_provider,
            base_currency=base_currency,
            tcc=tcc,
            initial_capital=initial_capital,
        )
        self.ome = {}
        self.order_to_instrument = {}
        self._channel = channel
        self._half_tick_size = {}
        self._fill_stop_order_at_price = accurate_stop_orders_execution
        if self._fill_stop_order_at_price:
            logger.info(f"[<y>{self.__class__.__name__}</y>] :: emulates stop orders executions at exact price")

    def get_orders(self, instrument: Instrument | None = None) -> dict[str, Order]:
        if instrument is not None:
            ome = self.ome.get(instrument)
            if ome is None:
                raise ValueError(f"ExchangeService:get_orders :: No OME configured for '{instrument}'!")

            return {o.id: o for o in ome.get_open_orders()}

        return {o.id: o for ome in self.ome.values() for o in ome.get_open_orders()}

    def get_position(self, instrument: Instrument) -> Position:
        if instrument in self.positions:
            return self.positions[instrument]

        # - initiolize OME for this instrument
        self.ome[instrument] = OrdersManagementEngine(
            instrument=instrument,
            time_provider=self.time_provider,
            tcc=self._tcc,  # type: ignore
            fill_stop_order_at_price=self._fill_stop_order_at_price,
        )

        # - initiolize empty position
        position = Position(instrument)  # type: ignore
        self._half_tick_size[instrument] = instrument.tick_size / 2  # type: ignore
        self.attach_positions(position)
        return self.positions[instrument]

    def update_position_price(self, time: dt_64, instrument: Instrument, price: float) -> None:
        super().update_position_price(time, instrument, price)

        # - first we need to update OME with new quote.
        # - if update is not a quote we need 'emulate' it.
        # - actually if SimulatedExchangeService is used in backtesting mode it will recieve only quotes
        # - case when we need that - SimulatedExchangeService is used for paper trading and data provider configured to listen to OHLC or TAS.
        # - probably we need to subscribe to quotes in real data provider in any case and then this emulation won't be needed.
        quote = price if isinstance(price, Quote) else self.emulate_quote_from_data(instrument, time, price)
        if quote is None:
            return

        # - process new quote
        self._process_new_quote(instrument, quote)

    def process_order(self, order: Order, update_locked_value: bool = True) -> None:
        _new = order.status == "NEW"
        _open = order.status == "OPEN"
        _cancel = order.status == "CANCELED"
        _closed = order.status == "CLOSED"
        if _new or _open:
            self.order_to_instrument[order.id] = order.instrument
        if (_cancel or _closed) and order.id in self.order_to_instrument:
            self.order_to_instrument.pop(order.id)
        return super().process_order(order, update_locked_value)

    def emulate_quote_from_data(
        self, instrument: Instrument, timestamp: dt_64, data: float | Timestamped
    ) -> Quote | None:
        if instrument not in self._half_tick_size:
            _ = self.get_position(instrument)

        if isinstance(data, Quote):
            return data

        elif isinstance(data, Trade):
            _ts2 = self._half_tick_size[instrument]
            if data.taker:  # type: ignore
                return Quote(timestamp, data.price - _ts2 * 2, data.price, 0, 0)  # type: ignore
            else:
                return Quote(timestamp, data.price, data.price + _ts2 * 2, 0, 0)  # type: ignore

        elif isinstance(data, Bar):
            _ts2 = self._half_tick_size[instrument]
            return Quote(timestamp, data.close - _ts2, data.close + _ts2, 0, 0)  # type: ignore

        elif isinstance(data, OrderBook):
            return data.to_quote()

        elif isinstance(data, float):
            _ts2 = self._half_tick_size[instrument]
            return Quote(timestamp, data - _ts2, data + _ts2, 0, 0)

        else:
            return None

    def _process_new_quote(self, instrument: Instrument, data: Quote) -> None:
        ome = self.ome.get(instrument)
        if ome is None:
            logger.warning("ExchangeService:update :: No OME configured for '{symbol}' yet !")
            return
        for r in ome.update_bbo(data):
            if r.exec is not None:
                self.order_to_instrument.pop(r.order.id)
                # - process methods will be called from stg context
                self._channel.send((instrument, "order", r.order, False))
                self._channel.send((instrument, "deals", [r.exec], False))



"""
Redis Streams Exporter for trading data.

This module provides an implementation of ITradeDataExport that exports trading data to Redis Streams.
"""

from typing import Any, Dict, List, Optional, TypeVar, cast

import redis
from redis.typing import EncodableT, FieldT

from qubx import logger
from qubx.core.basics import Instrument, Signal, TargetPosition, dt_64
from qubx.core.interfaces import IAccountViewer, ITradeDataExport
from qubx.exporters.formatters import DefaultFormatter, IExportFormatter


class RedisStreamsExporter(ITradeDataExport):
    """
    Exports trading data to Redis Streams.

    This exporter can be configured to export signals, target positions, and leverage changes.
    """

    def __init__(
        self,
        redis_url: str,
        strategy_name: str,
        export_signals: bool = True,
        export_targets: bool = True,
        export_position_changes: bool = True,
        signals_stream: Optional[str] = None,
        targets_stream: Optional[str] = None,
        position_changes_stream: Optional[str] = None,
        max_stream_length: int = 1000,
        formatter: Optional[IExportFormatter] = None,
    ):
        """
        Initialize the Redis Streams Exporter.

        Args:
            redis_url: Redis connection URL (e.g., "redis://localhost:6379/0")
            strategy_name: Name of the strategy (used in stream keys if not provided)
            export_signals: Whether to export signals
            export_targets: Whether to export target positions
            export_position_changes: Whether to export position changes
            signals_stream: Custom stream name for signals (default: "strategy:{strategy_name}:signals")
            targets_stream: Custom stream name for target positions (default: "strategy:{strategy_name}:targets")
            position_changes_stream: Custom stream name for position changes (default: "strategy:{strategy_name}:position_changes")
            max_stream_length: Maximum length of each stream
            formatter: Formatter to use for formatting data (default: DefaultFormatter)
        """
        self._redis = redis.from_url(redis_url)
        self._strategy_name = strategy_name

        self._export_signals = export_signals
        self._export_targets = export_targets
        self._export_position_changes = export_position_changes

        self._signals_stream = signals_stream or f"strategy:{strategy_name}:signals"
        self._targets_stream = targets_stream or f"strategy:{strategy_name}:targets"
        self._position_changes_stream = position_changes_stream or f"strategy:{strategy_name}:position_changes"

        self._max_stream_length = max_stream_length
        self._formatter = formatter or DefaultFormatter()

        self._instrument_to_previous_leverage = {}

        logger.info(
            f"[RedisStreamsExporter] Initialized for strategy '{strategy_name}' with "
            f"signals: {export_signals}, targets: {export_targets}, position_changes: {export_position_changes}"
        )

    def _prepare_for_redis(self, data: Dict[str, Any]) -> Dict[FieldT, EncodableT]:
        """
        Prepare data for Redis by ensuring all values are strings.

        Args:
            data: Dictionary with string keys and any values

        Returns:
            Dictionary with keys and values compatible with Redis
        """
        # Convert all values to strings and cast the result to the type expected by Redis
        string_dict = {k: str(v) for k, v in data.items()}
        return cast(Dict[FieldT, EncodableT], string_dict)

    def export_signals(self, time: dt_64, signals: List[Signal], account: IAccountViewer) -> None:
        """
        Export signals to Redis Stream.

        Args:
            time: Timestamp when the signals were generated
            signals: List of signals to export
            account: Account viewer to get account information like total capital, leverage, etc.
        """
        if not self._export_signals or not signals:
            return

        try:
            for signal in signals:
                # Format the signal using the formatter
                data = self._formatter.format_signal(time, signal, account)

                # Prepare data for Redis
                redis_data = self._prepare_for_redis(data)

                # Add to Redis stream
                self._redis.xadd(self._signals_stream, redis_data, maxlen=self._max_stream_length, approximate=True)

            logger.debug(f"[RedisStreamsExporter] Exported {len(signals)} signals to {self._signals_stream}")
        except Exception as e:
            logger.error(f"[RedisStreamsExporter] Failed to export signals: {e}")

    def export_target_positions(self, time: dt_64, targets: List[TargetPosition], account: IAccountViewer) -> None:
        """
        Export target positions to Redis Stream.

        Args:
            time: Timestamp when the target positions were generated
            targets: List of target positions to export
            account: Account viewer to get account information like total capital, leverage, etc.
        """
        if not self._export_targets or not targets:
            return

        try:
            for target in targets:
                # Format the target position using the formatter
                data = self._formatter.format_target_position(time, target, account)

                # Prepare data for Redis
                redis_data = self._prepare_for_redis(data)

                # Add to Redis stream
                self._redis.xadd(self._targets_stream, redis_data, maxlen=self._max_stream_length, approximate=True)

            logger.debug(f"[RedisStreamsExporter] Exported {len(targets)} target positions to {self._targets_stream}")
        except Exception as e:
            logger.error(f"[RedisStreamsExporter] Failed to export target positions: {e}")

    def export_position_changes(
        self, time: dt_64, instrument: Instrument, price: float, account: IAccountViewer
    ) -> None:
        """
        Export leverage changes to Redis Stream.

        Args:
            time: Timestamp when the leverage change occurred
            instrument: The instrument for which the leverage changed
            price: Price at which the leverage changed
            account: Account viewer to get account information like total capital, leverage, etc.
        """
        if not self._export_position_changes:
            return

        previous_leverage = self._instrument_to_previous_leverage.get(instrument)
        new_leverage = account.get_leverage(instrument)

        try:
            # Format the leverage change using the formatter
            data = self._formatter.format_position_change(time, instrument, price, account)

            # Prepare data for Redis
            redis_data = self._prepare_for_redis(data)

            # Add to Redis stream
            self._redis.xadd(
                self._position_changes_stream, redis_data, maxlen=self._max_stream_length, approximate=True
            )

            logger.debug(
                f"[RedisStreamsExporter] Exported position change for {instrument}: "
                f"{previous_leverage} -> {new_leverage} @ {price}"
            )
        except Exception as e:
            logger.error(f"[RedisStreamsExporter] Failed to export position change: {e}")



from typing import Any

from qubx.core.basics import Instrument, dt_64
from qubx.core.interfaces import IAccountViewer
from qubx.exporters.formatters.base import DefaultFormatter


class IncrementalFormatter(DefaultFormatter):
    """
    Incremental formatter for exporting trading data.
    """

    def format_position_change(
        self, time: dt_64, instrument: Instrument, price: float, account: IAccountViewer
    ) -> dict[str, Any]:
        return {}



import asyncio
import concurrent.futures
import re
from asyncio.exceptions import CancelledError
from collections import defaultdict
from threading import Thread
from types import FunctionType
from typing import Any, Awaitable, Callable, Dict, List, Optional, Set, Tuple

import pandas as pd

import ccxt.pro as cxp
from ccxt import (
    ExchangeClosedByUser,
    ExchangeError,
    ExchangeNotAvailable,
    NetworkError,
)
from ccxt.pro import Exchange
from qubx import logger
from qubx.core.basics import CtrlChannel, DataType, Instrument, ITimeProvider, dt_64
from qubx.core.helpers import BasicScheduler
from qubx.core.interfaces import IDataProvider
from qubx.core.series import Bar, Quote
from qubx.utils.misc import AsyncThreadLoop

from .exceptions import CcxtLiquidationParsingError, CcxtSymbolNotRecognized
from .utils import (
    ccxt_convert_funding_rate,
    ccxt_convert_liquidation,
    ccxt_convert_orderbook,
    ccxt_convert_ticker,
    ccxt_convert_trade,
    ccxt_find_instrument,
    instrument_to_ccxt_symbol,
)


class CcxtDataProvider(IDataProvider):
    time_provider: ITimeProvider
    _exchange: Exchange
    _scheduler: BasicScheduler | None = None

    # - subscriptions
    _subscriptions: Dict[str, Set[Instrument]]
    _sub_to_coro: Dict[str, concurrent.futures.Future]
    _sub_to_name: Dict[str, str]
    _sub_to_unsubscribe: Dict[str, Callable[[], Awaitable[None]]]
    _is_sub_name_enabled: Dict[str, bool]

    _sub_instr_to_time: Dict[Tuple[str, Instrument], dt_64]
    _last_quotes: Dict[Instrument, Optional[Quote]]
    _loop: AsyncThreadLoop
    _thread_event_loop: Thread
    _warmup_timeout: int

    _subscribers: Dict[str, Callable]
    _warmupers: Dict[str, Callable]

    def __init__(
        self,
        exchange: cxp.Exchange,
        time_provider: ITimeProvider,
        channel: CtrlChannel,
        max_ws_retries: int = 10,
        warmup_timeout: int = 120,
    ):
        self._exchange_id = str(exchange.name)
        self.time_provider = time_provider
        self.channel = channel
        self.max_ws_retries = max_ws_retries
        self._warmup_timeout = warmup_timeout

        # - create new even loop
        self._exchange = exchange
        self._loop = AsyncThreadLoop(self._exchange.asyncio_loop)

        self._last_quotes = defaultdict(lambda: None)
        self._subscriptions = defaultdict(set)
        self._sub_to_coro = {}
        self._sub_to_name = {}
        self._sub_to_unsubscribe = {}
        self._is_sub_name_enabled = defaultdict(lambda: False)
        self._symbol_to_instrument = {}
        self._subscribers = {
            n.split("_subscribe_")[1]: f
            for n, f in self.__class__.__dict__.items()
            if type(f) is FunctionType and n.startswith("_subscribe_")
        }
        self._warmupers = {
            n.split("_warmup_")[1]: f
            for n, f in self.__class__.__dict__.items()
            if type(f) is FunctionType and n.startswith("_warmup_")
        }
        logger.info(f"Initialized {self._exchange_id}")

    @property
    def is_simulation(self) -> bool:
        return False

    def subscribe(
        self,
        subscription_type: str,
        instruments: List[Instrument],
        reset: bool = False,
    ) -> None:
        _updated_instruments = set(instruments)
        # - update symbol to instrument mapping
        self._symbol_to_instrument.update({i.symbol: i for i in instruments})

        # - add existing subscription instruments if reset is False
        if not reset:
            _current_instruments = self.get_subscribed_instruments(subscription_type)
            _updated_instruments = _updated_instruments.union(_current_instruments)

        # - update subscriptions
        self._subscribe(_updated_instruments, subscription_type)

    def unsubscribe(self, subscription_type: str, instruments: List[Instrument]) -> None:
        # _current_instruments = self.get_subscribed_instruments(subscription_type)
        # _updated_instruments = set(_current_instruments).difference(instruments)
        # self._subscribe(_updated_instruments, subscription_type)
        # unsubscribe functionality is handled for ccxt via subscribe with reset=True
        pass

    def get_subscriptions(self, instrument: Instrument | None = None) -> List[str]:
        if instrument is not None:
            return [sub for sub, instrs in self._subscriptions.items() if instrument in instrs]
        return [sub for sub, instruments in self._subscriptions.items() if instruments]

    def get_subscribed_instruments(self, subscription_type: str | None = None) -> list[Instrument]:
        if not subscription_type:
            return list(self.subscribed_instruments)
        return list(self._subscriptions[subscription_type]) if subscription_type in self._subscriptions else []

    def has_subscription(self, instrument: Instrument, subscription_type: str) -> bool:
        sub = subscription_type.lower()
        return sub in self._subscriptions and instrument in self._subscriptions[sub]

    def warmup(self, warmups: Dict[Tuple[str, Instrument], str]) -> None:
        _coros = []

        for (sub_type, instrument), period in warmups.items():
            _sub_type, _params = DataType.from_str(sub_type)
            _warmuper = self._warmupers.get(_sub_type)
            if _warmuper is None:
                logger.warning(f"Warmup for {sub_type} is not supported")
                continue
            _coros.append(
                _warmuper(
                    self,
                    channel=self.channel,
                    instrument=instrument,
                    warmup_period=period,
                    **_params,
                )
            )

        async def gather_coros():
            return await asyncio.gather(*_coros)

        if _coros:
            self._loop.submit(gather_coros()).result(self._warmup_timeout)

    def get_quote(self, instrument: Instrument) -> Quote | None:
        return self._last_quotes[instrument]

    def get_ohlc(self, instrument: Instrument, timeframe: str, nbarsback: int) -> List[Bar]:
        assert nbarsback >= 1
        symbol = instrument.symbol
        since = self._time_msec_nbars_back(timeframe, nbarsback)

        # - retrieve OHLC data
        # - TODO: check if nbarsback > max_limit (1000) we need to do more requests
        # - TODO: how to get quoted volumes ?
        async def _get():
            return await self._exchange.fetch_ohlcv(
                symbol, self._get_exch_timeframe(timeframe), since=since, limit=nbarsback + 1
            )  # type: ignore

        res = self._loop.submit(_get()).result(60)

        _arr = []
        for oh in res:  # type: ignore
            _arr.append(
                Bar(oh[0] * 1_000_000, oh[1], oh[2], oh[3], oh[4], oh[6], oh[7])
                if len(oh) > 6
                else Bar(oh[0] * 1_000_000, oh[1], oh[2], oh[3], oh[4], oh[5])
            )

        return _arr

    def close(self):
        try:
            if hasattr(self._exchange, "close"):
                future = self._loop.submit(self._exchange.close())  # type: ignore
                # - wait for 5 seconds for connection to close
                future.result(5)
            else:
                del self._exchange
        except Exception as e:
            logger.error(e)

    @property
    def subscribed_instruments(self) -> Set[Instrument]:
        if not self._subscriptions:
            return set()
        return set.union(*self._subscriptions.values())

    @property
    def is_read_only(self) -> bool:
        _key = self._exchange.apiKey
        return _key is None or _key == ""

    def _subscribe(
        self,
        instruments: Set[Instrument],
        sub_type: str,
    ) -> None:
        _sub_type, _params = DataType.from_str(sub_type)
        _subscriber = self._subscribers.get(_sub_type)
        if _subscriber is None:
            raise ValueError(f"Subscription type {sub_type} is not supported")

        if sub_type in self._sub_to_coro:
            logger.debug(f"Canceling existing {sub_type} subscription for {self._subscriptions[sub_type]}")
            self._loop.submit(self._stop_subscriber(sub_type, self._sub_to_name[sub_type]))
            del self._sub_to_coro[sub_type]
            del self._sub_to_name[sub_type]

        if instruments is not None and len(instruments) == 0:
            return

        kwargs = {"instruments": instruments, **_params}
        _subscriber = self._subscribers[_sub_type]
        _subscriber_params = set(_subscriber.__code__.co_varnames[: _subscriber.__code__.co_argcount])
        # - get only parameters that are needed for subscriber
        kwargs = {k: v for k, v in kwargs.items() if k in _subscriber_params}
        self._sub_to_name[sub_type] = (name := self._get_subscription_name(_sub_type, **kwargs))
        self._sub_to_coro[sub_type] = self._loop.submit(_subscriber(self, name, _sub_type, self.channel, **kwargs))

        self._subscriptions[sub_type] = instruments

    def _time_msec_nbars_back(self, timeframe: str, nbarsback: int = 1) -> int:
        return (self.time_provider.time() - nbarsback * pd.Timedelta(timeframe)).asm8.item() // 1000000

    def _get_exch_timeframe(self, timeframe: str) -> str:
        if timeframe is not None:
            _t = re.match(r"(\d+)(\w+)", timeframe)
            timeframe = f"{_t[1]}{_t[2][0].lower()}" if _t and len(_t.groups()) > 1 else timeframe

        tframe = self._exchange.find_timeframe(timeframe)
        if tframe is None:
            raise ValueError(f"timeframe {timeframe} is not supported by {self._exchange.name}")

        return tframe

    def _get_exch_symbol(self, instrument: Instrument) -> str:
        return f"{instrument.base}/{instrument.quote}:{instrument.settle}"

    def _get_subscription_name(
        self, subscription: str, instruments: List[Instrument] | Set[Instrument] | Instrument | None = None, **kwargs
    ) -> str:
        if isinstance(instruments, Instrument):
            instruments = [instruments]
        _symbols = [instrument_to_ccxt_symbol(i) for i in instruments] if instruments is not None else []
        _name = f"{','.join(_symbols)} {subscription}" if _symbols else subscription
        if kwargs:
            kwargs_str = ",".join(f"{k}={v}" for k, v in kwargs.items())
            _name += f" ({kwargs_str})"
        return _name

    async def _stop_subscriber(self, sub_type: str, sub_name: str) -> None:
        try:
            self._is_sub_name_enabled[sub_name] = False  # stop the subscriber
            future = self._sub_to_coro[sub_type]
            total_sleep_time = 0.0
            while future.running():
                await asyncio.sleep(1.0)
                total_sleep_time += 1.0
                if total_sleep_time >= 20.0:
                    break

            if future.running():
                logger.warning(f"Subscriber {sub_name} is still running. Cancelling it.")
                future.cancel()
            else:
                logger.debug(f"Subscriber {sub_name} has been stopped")

            if sub_name in self._sub_to_unsubscribe:
                logger.debug(f"Unsubscribing from {sub_name}")
                await self._sub_to_unsubscribe[sub_name]()
                del self._sub_to_unsubscribe[sub_name]

            del self._is_sub_name_enabled[sub_name]
            logger.debug(f"Unsubscribed from {sub_name}")
        except Exception as e:
            logger.error(f"Error stopping {sub_name}")
            logger.exception(e)

    async def _listen_to_stream(
        self,
        subscriber: Callable[[], Awaitable[None]],
        exchange: Exchange,
        channel: CtrlChannel,
        name: str,
        unsubscriber: Callable[[], Awaitable[None]] | None = None,
    ):
        logger.info(f"Listening to {name}")
        if unsubscriber is not None:
            self._sub_to_unsubscribe[name] = unsubscriber

        self._is_sub_name_enabled[name] = True
        n_retry = 0
        while channel.control.is_set() and self._is_sub_name_enabled[name]:
            try:
                await subscriber()
                n_retry = 0
                if not self._is_sub_name_enabled[name]:
                    break
            except CcxtSymbolNotRecognized:
                continue
            except CancelledError:
                break
            except ExchangeClosedByUser:
                # - we closed connection so just stop it
                logger.info(f"{name} listening has been stopped")
                break
            except (NetworkError, ExchangeError, ExchangeNotAvailable) as e:
                logger.error(f"Error in {name} : {e}")
                await asyncio.sleep(1)
                continue
            except Exception as e:
                if not channel.control.is_set() or not self._is_sub_name_enabled[name]:
                    # If the channel is closed, then ignore all exceptions and exit
                    break
                logger.error(f"Exception in {name}")
                logger.exception(e)
                n_retry += 1
                if n_retry >= self.max_ws_retries:
                    logger.error(f"Max retries reached for {name}. Closing connection.")
                    del exchange
                    break
                await asyncio.sleep(min(2**n_retry, 60))  # Exponential backoff with a cap at 60 seconds

    #############################
    # - Warmup methods
    #############################
    async def _warmup_ohlc(
        self, channel: CtrlChannel, instrument: Instrument, warmup_period: str, timeframe: str
    ) -> None:
        nbarsback = pd.Timedelta(warmup_period) // pd.Timedelta(timeframe)
        exch_timeframe = self._get_exch_timeframe(timeframe)
        start = self._time_msec_nbars_back(timeframe, nbarsback)
        ohlcv = await self._exchange.fetch_ohlcv(instrument.symbol, exch_timeframe, since=start, limit=nbarsback + 1)
        logger.debug(f"{instrument}: loaded {len(ohlcv)} {timeframe} bars")
        channel.send(
            (
                instrument,
                DataType.OHLC[timeframe],
                [Bar(oh[0] * 1_000_000, oh[1], oh[2], oh[3], oh[4], oh[6], oh[7]) for oh in ohlcv],
                True,
            )
        )

    async def _warmup_trade(self, channel: CtrlChannel, instrument: Instrument, warmup_period: str):
        trades = await self._exchange.fetch_trades(instrument.symbol, since=self._time_msec_nbars_back(warmup_period))
        logger.debug(f"Loaded {len(trades)} trades for {instrument}")
        channel.send(
            (
                instrument,
                DataType.TRADE,
                [ccxt_convert_trade(trade) for trade in trades],
                True,
            )
        )

    def _call_by_market_type(
        self, subscriber: Callable[[list[Instrument]], Awaitable[None]], instruments: set[Instrument]
    ) -> Any:
        """Call subscriber for each market type"""
        _instr_by_type: dict[str, list[Instrument]] = defaultdict(list)
        for instr in instruments:
            _instr_by_type[instr.market_type].append(instr)

        # sort instruments by symbol
        for instrs in _instr_by_type.values():
            instrs.sort(key=lambda i: i.symbol)

        async def _call_subscriber():
            await asyncio.gather(*[subscriber(instrs) for instrs in _instr_by_type.values()])

        return _call_subscriber

    #############################
    # - Subscription methods
    #############################
    async def _subscribe_ohlc(
        self,
        name: str,
        sub_type: str,
        channel: CtrlChannel,
        instruments: Set[Instrument],
        timeframe: str = "1m",
    ):
        _instr_to_ccxt_symbol = {i: instrument_to_ccxt_symbol(i) for i in instruments}
        _exchange_timeframe = self._get_exch_timeframe(timeframe)
        _symbol_to_instrument = {_instr_to_ccxt_symbol[i]: i for i in instruments}

        async def watch_ohlcv(instruments: list[Instrument]):
            _symbol_timeframe_pairs = [[_instr_to_ccxt_symbol[i], _exchange_timeframe] for i in instruments]
            ohlcv = await self._exchange.watch_ohlcv_for_symbols(_symbol_timeframe_pairs)
            # - ohlcv is symbol -> timeframe -> list[timestamp, open, high, low, close, volume]
            for exch_symbol, _data in ohlcv.items():
                instrument = ccxt_find_instrument(exch_symbol, self._exchange, _symbol_to_instrument)
                for _, ohlcvs in _data.items():
                    for oh in ohlcvs:
                        channel.send(
                            (
                                instrument,
                                sub_type,
                                Bar(oh[0] * 1_000_000, oh[1], oh[2], oh[3], oh[4], oh[6], oh[7]),
                                False,  # not historical bar
                            )
                        )
                    if not (
                        self.has_subscription(instrument, DataType.ORDERBOOK)
                        or self.has_subscription(instrument, DataType.QUOTE)
                    ):
                        _price = ohlcvs[-1][4]
                        _s2 = instrument.tick_size / 2.0
                        _bid, _ask = _price - _s2, _price + _s2
                        self._last_quotes[instrument] = Quote(oh[0] * 1_000_000, _bid, _ask, 0.0, 0.0)

        # ohlc subscription reuses the same connection always, unsubscriptions don't work properly
        # but it's likely not very needed
        # async def un_watch_ohlcv(instruments: list[Instrument]):
        #     _symbol_timeframe_pairs = [[_instr_to_ccxt_symbol[i], _exchange_timeframe] for i in instruments]
        #     await self._exchange.un_watch_ohlcv_for_symbols(_symbol_timeframe_pairs)

        await self._listen_to_stream(
            subscriber=self._call_by_market_type(watch_ohlcv, instruments),
            exchange=self._exchange,
            channel=channel,
            name=name,
            # unsubscriber=self._call_by_market_type(un_watch_ohlcv, instruments),
        )

    async def _subscribe_trade(
        self,
        name: str,
        sub_type: str,
        channel: CtrlChannel,
        instruments: Set[Instrument],
    ):
        _instr_to_ccxt_symbol = {i: instrument_to_ccxt_symbol(i) for i in instruments}
        _symbol_to_instrument = {_instr_to_ccxt_symbol[i]: i for i in instruments}

        async def watch_trades(instruments: list[Instrument]):
            symbols = [_instr_to_ccxt_symbol[i] for i in instruments]
            trades = await self._exchange.watch_trades_for_symbols(symbols)
            exch_symbol = trades[0]["symbol"]
            instrument = ccxt_find_instrument(exch_symbol, self._exchange, _symbol_to_instrument)
            for trade in trades:
                channel.send((instrument, sub_type, ccxt_convert_trade(trade), False))

        async def un_watch_trades(instruments: list[Instrument]):
            symbols = [_instr_to_ccxt_symbol[i] for i in instruments]
            await self._exchange.un_watch_trades_for_symbols(symbols)

        await self._listen_to_stream(
            subscriber=self._call_by_market_type(watch_trades, instruments),
            exchange=self._exchange,
            channel=channel,
            name=name,
            unsubscriber=self._call_by_market_type(un_watch_trades, instruments),
        )

    async def _subscribe_orderbook(
        self,
        name: str,
        sub_type: str,
        channel: CtrlChannel,
        instruments: Set[Instrument],
    ):
        _instr_to_ccxt_symbol = {i: instrument_to_ccxt_symbol(i) for i in instruments}
        _symbol_to_instrument = {_instr_to_ccxt_symbol[i]: i for i in instruments}

        async def watch_orderbook(instruments: list[Instrument]):
            symbols = [_instr_to_ccxt_symbol[i] for i in instruments]
            ccxt_ob = await self._exchange.watch_order_book_for_symbols(symbols)
            exch_symbol = ccxt_ob["symbol"]
            instrument = ccxt_find_instrument(exch_symbol, self._exchange, _symbol_to_instrument)
            ob = ccxt_convert_orderbook(ccxt_ob, instrument)
            if ob is None:
                return
            quote = ob.to_quote()
            self._last_quotes[instrument] = quote
            channel.send((instrument, sub_type, ob, False))

        async def un_watch_orderbook(instruments: list[Instrument]):
            symbols = [_instr_to_ccxt_symbol[i] for i in instruments]
            await self._exchange.un_watch_order_book_for_symbols(symbols)

        await self._listen_to_stream(
            subscriber=self._call_by_market_type(watch_orderbook, instruments),
            exchange=self._exchange,
            channel=channel,
            name=name,
            unsubscriber=self._call_by_market_type(un_watch_orderbook, instruments),
        )

    async def _subscribe_quote(
        self,
        name: str,
        sub_type: str,
        channel: CtrlChannel,
        instruments: Set[Instrument],
    ):
        _instr_to_ccxt_symbol = {i: instrument_to_ccxt_symbol(i) for i in instruments}
        _symbol_to_instrument = {_instr_to_ccxt_symbol[i]: i for i in instruments}

        async def watch_quote(instruments: list[Instrument]):
            symbols = [_instr_to_ccxt_symbol[i] for i in instruments]
            ccxt_tickers: dict[str, dict] = await self._exchange.watch_tickers(symbols)
            for exch_symbol, ccxt_ticker in ccxt_tickers.items():
                instrument = ccxt_find_instrument(exch_symbol, self._exchange, _symbol_to_instrument)
                quote = ccxt_convert_ticker(ccxt_ticker, instrument)
                self._last_quotes[instrument] = quote
                channel.send((instrument, sub_type, quote, False))

        async def un_watch_quote(instruments: list[Instrument]):
            symbols = [_instr_to_ccxt_symbol[i] for i in instruments]
            await self._exchange.un_watch_tickers(symbols)

        await self._listen_to_stream(
            subscriber=self._call_by_market_type(watch_quote, instruments),
            exchange=self._exchange,
            channel=channel,
            name=name,
            unsubscriber=self._call_by_market_type(un_watch_quote, instruments),
        )

    async def _subscribe_liquidation(
        self,
        name: str,
        sub_type: str,
        channel: CtrlChannel,
        instruments: Set[Instrument],
    ):
        _instr_to_ccxt_symbol = {i: instrument_to_ccxt_symbol(i) for i in instruments}
        _symbol_to_instrument = {_instr_to_ccxt_symbol[i]: i for i in instruments}

        async def watch_liquidation(instruments: list[Instrument]):
            symbols = [_instr_to_ccxt_symbol[i] for i in instruments]
            liquidations = await self._exchange.watch_liquidations_for_symbols(symbols)
            for liquidation in liquidations:
                try:
                    instrument = ccxt_find_instrument(liquidation["symbol"], self._exchange, _symbol_to_instrument)
                    channel.send((instrument, sub_type, ccxt_convert_liquidation(liquidation), False))
                except CcxtLiquidationParsingError:
                    logger.debug(f"Could not parse liquidation {liquidation}")
                    continue

        async def un_watch_liquidation(instruments: list[Instrument]):
            symbols = [_instr_to_ccxt_symbol[i] for i in instruments]
            unwatch = getattr(self._exchange, "un_watch_liquidations_for_symbols", lambda _: None)(symbols)
            if unwatch is not None:
                await unwatch

        # - fetching of liquidations for warmup is not supported by ccxt
        await self._listen_to_stream(
            subscriber=self._call_by_market_type(watch_liquidation, instruments),
            exchange=self._exchange,
            channel=channel,
            name=name,
            unsubscriber=self._call_by_market_type(un_watch_liquidation, instruments),
        )

    async def _subscribe_funding_rate(
        self,
        name: str,
        sub_type: str,
        channel: CtrlChannel,
    ):
        # it is expected that we can retrieve funding rates for all instruments
        async def watch_funding_rates():
            funding_rates = await self._exchange.watch_funding_rates()  # type: ignore
            instrument_to_funding_rate = {}
            for symbol, info in funding_rates.items():
                try:
                    instrument = ccxt_find_instrument(symbol, self._exchange)
                    instrument_to_funding_rate[instrument] = ccxt_convert_funding_rate(info)
                except CcxtSymbolNotRecognized:
                    continue
            channel.send((None, sub_type, instrument_to_funding_rate, False))

        async def un_watch_funding_rates():
            unwatch = getattr(self._exchange, "un_watch_funding_rates", lambda: None)()
            if unwatch is not None:
                await unwatch

        await self._listen_to_stream(
            subscriber=watch_funding_rates,
            exchange=self._exchange,
            channel=channel,
            name=name,
            unsubscriber=un_watch_funding_rates,
        )

    def exchange(self) -> str:
        return self._exchange_id.upper()



__all__ = [
    "srows",
    "scols",
    "continuous_periods",
    "ohlc_resample",
    "retain_columns_and_join",
    "select_column_and_join",
    "dict_to_frame",
    "drop_duplicated_indexes",
    "process_duplicated_indexes",
]

from .utils import (
    continuous_periods,
    dict_to_frame,
    drop_duplicated_indexes,
    ohlc_resample,
    process_duplicated_indexes,
    retain_columns_and_join,
    scols,
    select_column_and_join,
    srows,
)



import traceback
from threading import Thread
from typing import Any, Callable

from qubx import logger
from qubx.core.basics import (
    SW,
    AssetBalance,
    CtrlChannel,
    DataType,
    Instrument,
    ITimeProvider,
    MarketType,
    Order,
    OrderRequest,
    Position,
    dt_64,
)
from qubx.core.helpers import (
    BasicScheduler,
    CachedMarketDataHolder,
    set_parameters_to_object,
)
from qubx.core.interfaces import (
    IAccountProcessor,
    IBroker,
    IDataProvider,
    IMarketManager,
    IPositionGathering,
    IProcessingManager,
    IStrategy,
    IStrategyContext,
    ISubscriptionManager,
    ITradeDataExport,
    ITradingManager,
    IUniverseManager,
    PositionsTracker,
    RemovalPolicy,
)
from qubx.core.loggers import StrategyLogging
from qubx.data.readers import DataReader
from qubx.gathering.simplest import SimplePositionGatherer
from qubx.trackers.sizers import FixedSizer

from .mixins import (
    MarketManager,
    ProcessingManager,
    SubscriptionManager,
    TradingManager,
    UniverseManager,
)


class StrategyContext(IStrategyContext):
    DEFAULT_POSITION_TRACKER: Callable[[], PositionsTracker] = lambda: PositionsTracker(
        FixedSizer(1.0, amount_in_quote=False)
    )

    _market_data_provider: IMarketManager
    _universe_manager: IUniverseManager
    _subscription_manager: ISubscriptionManager
    _trading_manager: ITradingManager
    _processing_manager: IProcessingManager

    _broker: IBroker  # service for exchange API: orders managemewnt
    _logging: StrategyLogging  # recording all activities for the strat: execs, positions, portfolio
    _data_provider: IDataProvider  # market data provider
    _cache: CachedMarketDataHolder
    _scheduler: BasicScheduler
    _initial_instruments: list[Instrument]

    _thread_data_loop: Thread | None = None  # market data loop
    _is_initialized: bool = False
    _exporter: ITradeDataExport | None = None  # Add exporter attribute

    def __init__(
        self,
        strategy: IStrategy,
        broker: IBroker,
        data_provider: IDataProvider,
        account: IAccountProcessor,
        scheduler: BasicScheduler,
        time_provider: ITimeProvider,
        instruments: list[Instrument],
        logging: StrategyLogging,
        config: dict[str, Any] | None = None,
        position_gathering: IPositionGathering | None = None,  # TODO: make position gathering part of the strategy
        aux_data_provider: DataReader | None = None,
        exporter: ITradeDataExport | None = None,  # Add exporter parameter
    ) -> None:
        self.account = account
        self.strategy = self.__instantiate_strategy(strategy, config)

        self._time_provider = time_provider
        self._broker = broker
        self._data_provider = data_provider
        self._logging = logging
        self._scheduler = scheduler
        self._initial_instruments = instruments

        self._cache = CachedMarketDataHolder()
        self._exporter = exporter  # Store the exporter

        __position_tracker = self.strategy.tracker(self)
        if __position_tracker is None:
            __position_tracker = StrategyContext.DEFAULT_POSITION_TRACKER()

        __position_gathering = position_gathering if position_gathering is not None else SimplePositionGatherer()

        self._subscription_manager = SubscriptionManager(
            data_provider=self._data_provider,
            default_base_subscription=DataType.ORDERBOOK if not self._data_provider.is_simulation else DataType.NONE,
        )
        self.account.set_subscription_manager(self._subscription_manager)

        self._market_data_provider = MarketManager(
            time_provider=self._time_provider,
            cache=self._cache,
            data_provider=self._data_provider,
            universe_manager=self,
            aux_data_provider=aux_data_provider,
        )
        self._universe_manager = UniverseManager(
            context=self,
            strategy=self.strategy,
            broker=self._data_provider,
            trading_service=self._broker,
            cache=self._cache,
            logging=self._logging,
            subscription_manager=self,
            trading_manager=self,
            time_provider=self,
            account=self.account,
            position_gathering=__position_gathering,
        )
        self._trading_manager = TradingManager(
            time_provider=self,
            broker=self._broker,
            account=self.account,
            strategy_name=self.strategy.__class__.__name__,
        )
        self._processing_manager = ProcessingManager(
            context=self,
            strategy=self.strategy,
            logging=self._logging,
            market_data=self,
            subscription_manager=self,
            time_provider=self,
            account=self.account,
            position_tracker=__position_tracker,
            position_gathering=__position_gathering,
            universe_manager=self._universe_manager,
            cache=self._cache,
            scheduler=self._scheduler,
            is_simulation=self._data_provider.is_simulation,
            exporter=self._exporter,  # Pass exporter to processing manager
        )
        self.__post_init__()

    def __post_init__(self) -> None:
        self.strategy.on_init(self)
        # - update cache default timeframe
        sub_type = self.get_base_subscription()
        _, params = DataType.from_str(sub_type)
        __default_timeframe = params.get("timeframe", "1sec")
        self._cache.update_default_timeframe(__default_timeframe)

    def start(self, blocking: bool = False):
        if self._is_initialized:
            raise ValueError("Strategy is already started !")

        # - run cron scheduler
        self._scheduler.run()

        # - create incoming market data processing
        databus = self._data_provider.channel
        databus.register(self)

        # - start account processing
        self.account.start()

        # - update universe with initial instruments after the strategy is initialized
        self.set_universe(self._initial_instruments, skip_callback=True)

        # - initialize strategy (should we do that after any first market data received ?)
        if not self._is_initialized:
            try:
                self.strategy.on_start(self)
                self._is_initialized = True
            except Exception as strat_error:
                logger.error(
                    f"[StrategyContext] :: Strategy {self.strategy.__class__.__name__} raised an exception in on_start: {strat_error}"
                )
                logger.error(traceback.format_exc())
                return

        # - for live we run loop
        if not self._data_provider.is_simulation:
            self._thread_data_loop = Thread(target=self.__process_incoming_data_loop, args=(databus,), daemon=True)
            self._thread_data_loop.start()
            logger.info("[StrategyContext] :: strategy is started in thread")
            if blocking:
                self._thread_data_loop.join()

    def stop(self):
        # - invoke strategy's stop code
        try:
            self.strategy.on_stop(self)
        except Exception as strat_error:
            logger.error(
                f"[<y>StrategyContext</y>] :: Strategy {self.strategy.__class__.__name__} raised an exception in on_stop: {strat_error}"
            )
            logger.opt(colors=False).error(traceback.format_exc())

        if self._thread_data_loop:
            self._data_provider.close()
            self._data_provider.channel.stop()
            self._thread_data_loop.join()
            self._thread_data_loop = None

        # - stop account processing
        self.account.stop()

        # - close logging
        self._logging.close()

    def is_running(self):
        return self._thread_data_loop is not None and self._thread_data_loop.is_alive()

    @property
    def is_simulation(self) -> bool:
        return self._data_provider.is_simulation

    # IAccountViewer delegation

    # capital information
    def get_capital(self) -> float:
        return self.account.get_capital()

    def get_total_capital(self) -> float:
        return self.account.get_total_capital()

    # balance and position information
    def get_balances(self) -> dict[str, AssetBalance]:
        return dict(self.account.get_balances())

    def get_positions(self) -> dict[Instrument, Position]:
        return dict(self.account.get_positions())

    def get_position(self, instrument: Instrument) -> Position:
        return self.account.get_position(instrument)

    @property
    def positions(self):
        return self.account.positions

    def get_orders(self, instrument: Instrument | None = None) -> dict[str, Order]:
        return self.account.get_orders(instrument)

    def position_report(self) -> dict:
        return self.account.position_report()

    # leverage information
    def get_leverage(self, instrument: Instrument) -> float:
        return self.account.get_leverage(instrument)

    def get_leverages(self) -> dict[Instrument, float]:
        return self.account.get_leverages()

    def get_net_leverage(self) -> float:
        return self.account.get_net_leverage()

    def get_gross_leverage(self) -> float:
        return self.account.get_gross_leverage()

    # margin information
    def get_total_required_margin(self) -> float:
        return self.account.get_total_required_margin()

    def get_available_margin(self) -> float:
        return self.account.get_available_margin()

    def get_margin_ratio(self) -> float:
        return self.account.get_margin_ratio()

    # IMarketDataProvider delegation
    def time(self) -> dt_64:
        return self._market_data_provider.time()

    def ohlc(self, instrument: Instrument, timeframe: str | None = None, length: int | None = None):
        return self._market_data_provider.ohlc(instrument, timeframe, length)

    def quote(self, instrument: Instrument):
        return self._market_data_provider.quote(instrument)

    def get_data(self, instrument: Instrument, sub_type: str) -> list[Any]:
        return self._market_data_provider.get_data(instrument, sub_type)

    def get_aux_data(self, data_id: str, **parameters):
        return self._market_data_provider.get_aux_data(data_id, **parameters)

    def get_instruments(self):
        return self._market_data_provider.get_instruments()

    def query_instrument(self, symbol: str, exchange: str | None = None) -> Instrument | None:
        return self._market_data_provider.query_instrument(
            symbol, exchange if exchange is not None else self.exchanges[0]
        )

    # ITradingManager delegation
    def trade(self, instrument: Instrument, amount: float, price: float | None = None, time_in_force="gtc", **options):
        return self._trading_manager.trade(instrument, amount, price, time_in_force, **options)

    def submit_orders(self, order_requests: list[OrderRequest]) -> list[Order]:
        return self._trading_manager.submit_orders(order_requests)

    def set_target_position(
        self, instrument: Instrument, target: float, price: float | None = None, **options
    ) -> Order:
        return self._trading_manager.set_target_position(instrument, target, price, **options)

    def close_position(self, instrument: Instrument) -> None:
        return self._trading_manager.close_position(instrument)

    def close_positions(self, market_type: MarketType | None = None) -> None:
        return self._trading_manager.close_positions(market_type)

    def cancel_order(self, order_id: str) -> None:
        return self._trading_manager.cancel_order(order_id)

    def cancel_orders(self, instrument: Instrument):
        return self._trading_manager.cancel_orders(instrument)

    # IUniverseManager delegation
    def set_universe(
        self, instruments: list[Instrument], skip_callback: bool = False, if_has_position_then: RemovalPolicy = "close"
    ):
        return self._universe_manager.set_universe(instruments, skip_callback, if_has_position_then)

    def add_instruments(self, instruments: list[Instrument]):
        return self._universe_manager.add_instruments(instruments)

    def remove_instruments(self, instruments: list[Instrument]):
        return self._universe_manager.remove_instruments(instruments)

    @property
    def instruments(self):
        return self._universe_manager.instruments

    @property
    def exchanges(self) -> list[str]:
        return self._trading_manager.exchanges()

    # ISubscriptionManager delegation
    def subscribe(self, subscription_type: str, instruments: list[Instrument] | Instrument | None = None):
        return self._subscription_manager.subscribe(subscription_type, instruments)

    def unsubscribe(self, subscription_type: str, instruments: list[Instrument] | Instrument | None = None):
        return self._subscription_manager.unsubscribe(subscription_type, instruments)

    def has_subscription(self, instrument: Instrument, subscription_type: str):
        return self._subscription_manager.has_subscription(instrument, subscription_type)

    def get_subscriptions(self, instrument: Instrument | None = None) -> list[str]:
        return self._subscription_manager.get_subscriptions(instrument)

    def get_base_subscription(self) -> str:
        return self._subscription_manager.get_base_subscription()

    def set_base_subscription(self, subscription_type: str):
        return self._subscription_manager.set_base_subscription(subscription_type)

    def get_subscribed_instruments(self, subscription_type: str | None = None) -> list[Instrument]:
        return self._subscription_manager.get_subscribed_instruments(subscription_type)

    def get_warmup(self, subscription_type: str) -> str | None:
        return self._subscription_manager.get_warmup(subscription_type)

    def set_warmup(self, configs: dict[Any, str]):
        return self._subscription_manager.set_warmup(configs)

    def commit(self):
        return self._subscription_manager.commit()

    @property
    def auto_subscribe(self) -> bool:
        return self._subscription_manager.auto_subscribe

    @auto_subscribe.setter
    def auto_subscribe(self, value: bool):
        self._subscription_manager.auto_subscribe = value

    # IProcessingManager delegation
    def process_data(self, instrument: Instrument, d_type: str, data: Any, is_historical: bool):
        return self._processing_manager.process_data(instrument, d_type, data, is_historical)

    def set_fit_schedule(self, schedule: str):
        return self._processing_manager.set_fit_schedule(schedule)

    def set_event_schedule(self, schedule: str):
        return self._processing_manager.set_event_schedule(schedule)

    def get_event_schedule(self, event_id: str) -> str | None:
        return self._processing_manager.get_event_schedule(event_id)

    def is_fitted(self) -> bool:
        return self._processing_manager.is_fitted()

    # private methods
    def __process_incoming_data_loop(self, channel: CtrlChannel):
        logger.info("[StrategyContext] :: Start processing market data")
        while channel.control.is_set():
            with SW("StrategyContext._process_incoming_data"):
                # - waiting for incoming market data
                instrument, d_type, data, hist = channel.receive()
                if self.process_data(instrument, d_type, data, hist):
                    channel.stop()
                    break
        logger.info("[StrategyContext] :: Market data processing stopped")

    def __instantiate_strategy(self, strategy: IStrategy, config: dict[str, Any] | None) -> IStrategy:
        __strategy = strategy() if isinstance(strategy, type) else strategy
        __strategy.ctx = self
        set_parameters_to_object(__strategy, **config if config else {})
        return __strategy
